# Functions that don't fit anywhere else.

def parse_class_name(uri: str):
    """Parse the class name in a URI.

    The input is expected to be of the form 'http://sbols.org/v3#Component'
    or 'http://example.com/ApplicationSpecificClass'. This function would
    return 'Component' and 'ApplicationSpecificClass' respectively.
    """
    if '#' in uri:
        return uri[uri.rindex('#')+1:]
    elif '/' in uri:
        return uri[uri.rindex('/')+1:]
    raise ValueError(f'Cannot parse class name from {uri}. URI must use either / '
                     'or # as a delimiter.')



import math
from typing import List

from . import *


class Interaction(Identified):

    def __init__(self, interaction_types: List[str],
                 *, identity: str = None,
                 type_uri: str = SBOL_INTERACTION) -> None:
        super().__init__(identity, type_uri)
        self.types = URIProperty(self, SBOL_TYPE, 1, math.inf,
                                 initial_value=interaction_types)
        self.participations = OwnedObject(self, SBOL_PARTICIPATIONS, 0, math.inf,
                                          type_constraint=Participation)

    def validate(self) -> None:
        super().validate()


def build_interaction(identity: str, *, type_uri: str = SBOL_INTERACTION) -> SBOLObject:
    interaction_type = PYSBOL3_MISSING
    obj = Interaction([interaction_type], identity=identity, type_uri=type_uri)
    # Remove the dummy type
    obj._properties[SBOL_TYPE] = []
    return obj


Document.register_builder(SBOL_INTERACTION, build_interaction)



import unittest

import rdflib

import sbol3


class TestInteraction(unittest.TestCase):

    def test_create(self):
        types = [sbol3.SBO_INHIBITION]
        interaction = sbol3.Interaction(types)
        self.assertIsNotNone(interaction)
        self.assertEqual(types, interaction.types)

    def test_read(self):
        component_id = 'https://github.com/synbiodex/pysbol3/component1'
        interaction_id = 'https://github.com/synbiodex/pysbol3/interaction1'
        rdf_type = rdflib.RDF.type
        nt_data = f'<{component_id}> <{rdf_type}> <{sbol3.SBOL_COMPONENT}> .\n'
        nt_data += f'<{component_id}> <{sbol3.SBOL_TYPE}> <{sbol3.SBO_DNA}> .\n'
        nt_data += f'<{component_id}> <{sbol3.SBOL_INTERACTIONS}> <{interaction_id}> .\n'
        nt_data += f'<{interaction_id}> <{rdf_type}> <{sbol3.SBOL_INTERACTION}> .\n'
        nt_data += f'<{interaction_id}> <{sbol3.SBOL_TYPE}> <{sbol3.SBO_INHIBITION}> .\n'
        doc = sbol3.Document()
        doc.read_string(nt_data, 'ttl')
        component = doc.find(component_id)
        self.assertIsNotNone(component)
        self.assertIsInstance(component, sbol3.Component)
        interaction = component.interactions[0]
        self.assertEqual(interaction_id, interaction.identity)
        self.assertEqual([sbol3.SBO_INHIBITION], interaction.types)


if __name__ == '__main__':
    unittest.main()



import unittest

import sbol3


class TestRange(unittest.TestCase):

    def test_creation(self):
        start = 1
        end = 10
        r = sbol3.Range(sbol3.PYSBOL3_MISSING, start, end)
        self.assertIsNotNone(r)
        self.assertEqual(start, r.start)
        self.assertEqual(end, r.end)

    def test_invalid_create(self):
        start = 0
        end = 10
        with self.assertRaises(sbol3.ValidationError):
            sbol3.Range(sbol3.PYSBOL3_MISSING, start, end)
        start = 1
        end = 0
        with self.assertRaises(sbol3.ValidationError):
            sbol3.Range(sbol3.PYSBOL3_MISSING, start, end)
        # end must be >= start
        start = 10
        end = 9
        with self.assertRaises(sbol3.ValidationError):
            sbol3.Range(sbol3.PYSBOL3_MISSING, start, end)
        start = 7
        end = 7
        r = sbol3.Range(sbol3.PYSBOL3_MISSING, start, end)
        self.assertEqual(start, r.start)
        self.assertEqual(end, r.end)


class TestCut(unittest.TestCase):

    def test_creation(self):
        at = 1
        cut = sbol3.Cut(sbol3.PYSBOL3_MISSING, at)
        self.assertIsNotNone(cut)
        self.assertEqual(at, cut.at)
        at = 0
        cut = sbol3.Cut(sbol3.PYSBOL3_MISSING, at)
        self.assertIsNotNone(cut)
        self.assertEqual(at, cut.at)

    def test_invalid_create(self):
        # At must be >= 0
        at = -1
        with self.assertRaises(sbol3.ValidationError):
            sbol3.Cut(sbol3.PYSBOL3_MISSING, at)


class TestEntireSequence(unittest.TestCase):

    def test_creation(self):
        # EntireSequence has no properties, so there isn't much to test here
        es = sbol3.EntireSequence(sbol3.PYSBOL3_MISSING)
        self.assertIsNotNone(es)


if __name__ == '__main__':
    unittest.main()



import os
import unittest

import sbol3

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
SBOL3_LOCATION = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL3')


class TestMeasure(unittest.TestCase):

    def test_create(self):
        unit = 'https://sbolstandard.org/examples/millimolePerLitre'
        value = 0.1
        types = ['https://identifiers.org/SBO:0000196',
                 'https://identifiers.org/SBO:0000197']
        measure = sbol3.Measure(value, unit)
        measure.types = types
        self.assertIsNotNone(measure)
        self.assertIsInstance(measure, sbol3.Measure)
        self.assertEqual(value, measure.value)
        self.assertCountEqual(types, measure.types)
        self.assertEqual(unit, measure.unit)

    def test_read_from_file(self):
        test_file = os.path.join(SBOL3_LOCATION, 'measurement_entity',
                                 'measurement', 'measurement.nt')
        doc = sbol3.Document()
        doc.read(test_file, sbol3.NTRIPLES)
        uri = 'https://sbolstandard.org/examples/M9_Glucose_CAA/CaCl2/measure1'
        measure = doc.find(uri)
        self.assertIsNotNone(measure)
        self.assertIsInstance(measure, sbol3.Measure)
        self.assertEqual(0.1, measure.value)
        self.assertCountEqual(['https://identifiers.org/SBO:0000196',
                               'https://identifiers.org/SBO:0000197'],
                              measure.types)
        self.assertEqual('https://sbolstandard.org/examples/millimolePerLitre',
                         measure.unit)
        self.assertEqual('measure1', measure.display_id)


class TestPrefixedUnit(unittest.TestCase):

    def test_create(self):
        display_id = 'millimole'
        symbol = display_id
        label = display_id
        prefix = 'https://sbolstandard.org/examples/milli'
        unit = 'https://sbolstandard.org/examples/mole'
        punit = sbol3.PrefixedUnit(display_id, symbol, label, unit, prefix)
        self.assertIsNotNone(punit)
        self.assertIsInstance(punit, sbol3.PrefixedUnit)
        self.assertCountEqual(prefix, punit.prefix)
        self.assertEqual(unit, punit.unit)
        self.assertEqual(symbol, punit.symbol)
        self.assertEqual(label, punit.label)
        self.assertEqual(display_id, punit.display_id)

    def test_read_from_file(self):
        test_file = os.path.join(SBOL3_LOCATION, 'measurement_entity',
                                 'measurement', 'measurement.nt')
        doc = sbol3.Document()
        doc.read(test_file, sbol3.NTRIPLES)
        uri = 'https://sbolstandard.org/examples/millimole'
        punit = doc.find(uri)
        self.assertIsNotNone(punit)
        self.assertIsInstance(punit, sbol3.PrefixedUnit)
        self.assertCountEqual('https://sbolstandard.org/examples/milli', punit.prefix)
        self.assertEqual('https://sbolstandard.org/examples/mole', punit.unit)
        self.assertEqual('millimole', punit.label)
        self.assertEqual('millimole', punit.name)
        self.assertEqual('millimole', punit.display_id)


class TestSingularUnit(unittest.TestCase):

    def test_create(self):
        display_id = 'litre'
        symbol = display_id
        label = display_id
        unit = 'https://sbolstandard.org/examples/litre'
        factor = 0.001
        sunit = sbol3.SingularUnit(display_id, symbol, label)
        sunit.unit = unit
        sunit.factor = factor
        self.assertIsNotNone(sunit)
        self.assertIsInstance(sunit, sbol3.SingularUnit)
        self.assertEqual(0.001, sunit.factor)
        self.assertEqual(unit, sunit.unit)
        self.assertEqual(symbol, sunit.symbol)
        self.assertEqual(label, sunit.label)

    def test_read_from_file(self):
        test_file = os.path.join(SBOL3_LOCATION, 'measurement_entity',
                                 'measurement', 'measurement.nt')
        doc = sbol3.Document()
        doc.read(test_file, sbol3.NTRIPLES)
        uri = 'https://sbolstandard.org/examples/litre'
        sunit = doc.find(uri)
        self.assertIsNotNone(sunit)
        self.assertIsInstance(sunit, sbol3.SingularUnit)
        self.assertEqual(0.001, sunit.factor)
        self.assertIsNone(sunit.unit)
        self.assertIsNone(sunit.symbol)
        self.assertCountEqual(['L', 'L2'], sunit.alternative_symbols)
        self.assertEqual('liter', sunit.label)
        self.assertCountEqual(['liter', 'litre2'], sunit.alternative_labels)
        self.assertIsNotNone(sunit.comment)
        self.assertIsNotNone(sunit.long_comment)
        self.assertEqual('liter', sunit.name)
        self.assertIsNotNone(sunit.description)
        self.assertEqual('litre', sunit.display_id)


if __name__ == '__main__':
    unittest.main()



import math
import os
import tempfile
import unittest

import rdflib

import sbol3


PYSBOL3_CUSTOM_TOP = 'https://github.com/synbiodex/pysbol3#customTop'
PYSBOL3_CUSTOM_BOOL = 'https://github.com/synbiodex/pysbol3#customBool'
PYSBOL3_CUSTOM_CHILD = 'https://github.com/synbiodex/pysbol3#customChildren'
PYSBOL3_CUSTOM_IDENTIFIED = 'https://github.com/synbiodex/pysbol3#customIdentified'
PYSBOL3_CUSTOM_INT = 'https://github.com/synbiodex/pysbol3#customInt'


class CustomTopClass(sbol3.CustomTopLevel):
    def __init__(self, identity, type_uri=PYSBOL3_CUSTOM_TOP):
        super().__init__(identity, type_uri)
        # Also test the boolean list while we're here
        self.foo_bool = sbol3.BooleanProperty(self, PYSBOL3_CUSTOM_BOOL,
                                              0, math.inf)
        self.children = sbol3.OwnedObject(self, PYSBOL3_CUSTOM_CHILD, 0, math.inf)


class CustomIdentifiedClass(sbol3.CustomIdentified):
    def __init__(self, type_uri=PYSBOL3_CUSTOM_IDENTIFIED, identity=None):
        super().__init__(type_uri, identity=identity)
        # Also test the int list while we're here
        self.foo_int = sbol3.IntProperty(self, PYSBOL3_CUSTOM_INT,
                                         0, math.inf)


class TestCustomTopLevel(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        sbol3.Document.register_builder(PYSBOL3_CUSTOM_TOP, CustomTopClass)

    @classmethod
    def tearDownClass(cls) -> None:
        # Go behind the scenes to clean up
        # TODO: should there be a way to deregister a builder?
        store = sbol3.Document._uri_type_map
        if PYSBOL3_CUSTOM_TOP in store:
            del store[PYSBOL3_CUSTOM_TOP]

    def test_create(self):
        custom_type = 'https://github.com/synbiodex/pysbol3/CustomType'
        ctl = sbol3.CustomTopLevel('custom1', custom_type)
        self.assertEqual(custom_type, ctl.type_uri)
        # Go behind the scenes to verify
        self.assertEqual(rdflib.URIRef(sbol3.SBOL_TOP_LEVEL),
                         ctl._properties[rdflib.RDF.type][0])

    # TODO: We really want to verify the serialization of the custom top
    #       level as well. There is no Document.writeString() yet.

    def test_round_trip(self):
        # Test the boolean list property, which is not used by the
        # core SBOL 3 data model
        obj_name = 'bool_test'
        obj = CustomTopClass(obj_name)
        self.assertEqual([], obj.foo_bool)
        obj.foo_bool.append(True)
        obj.foo_bool.append(False)
        self.assertEqual([True, False], obj.foo_bool)
        doc = sbol3.Document()
        doc.add(obj)
        doc2 = sbol3.Document()
        # Round trip the document
        with tempfile.TemporaryDirectory() as tmpdirname:
            test_file = os.path.join(tmpdirname, 'custom.nt')
            doc.write(test_file, sbol3.NTRIPLES)
            doc2.read(test_file, sbol3.NTRIPLES)
        obj2 = doc2.find(obj_name)
        # The lists are necessarily unordered because of RDF
        # Compare specially
        self.assertCountEqual([True, False], obj2.foo_bool)

    def test_none_identity(self):
        # Make sure a ValueError is raised if None is passed
        # as a CustomTopLevel identity. And also if identity
        # is an empty string or not a string.
        with self.assertRaises(ValueError):
            obj = CustomTopClass(None)
        with self.assertRaises(ValueError):
            obj = CustomTopClass('')
        with self.assertRaises(ValueError):
            obj = CustomTopClass(3)


class TestCustomIdentified(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        sbol3.Document.register_builder(PYSBOL3_CUSTOM_TOP, CustomTopClass)
        sbol3.Document.register_builder(PYSBOL3_CUSTOM_IDENTIFIED, CustomIdentifiedClass)

    @classmethod
    def tearDownClass(cls) -> None:
        # Go behind the scenes to clean up
        # TODO: should there be a way to deregister a builder?
        store = sbol3.Document._uri_type_map
        for uri in [PYSBOL3_CUSTOM_TOP, PYSBOL3_CUSTOM_IDENTIFIED]:
            if uri in store:
                del store[uri]

    def test_create(self):
        custom_type = 'https://github.com/synbiodex/pysbol3/CustomType'
        ctl = sbol3.CustomIdentified(identity='custom1',
                                     type_uri=custom_type)
        self.assertEqual(custom_type, ctl.type_uri)
        # Go behind the scenes to verify
        self.assertEqual(rdflib.URIRef(sbol3.SBOL_IDENTIFIED),
                         ctl._properties[rdflib.RDF.type][0])

    # TODO: We really want to verify the serialization of the custom
    #       identified as well. We need to attach it to a top level
    #       to see that work.

    def test_round_trip(self):
        # Test the int list property, which is not used by the
        # core SBOL 3 data model
        obj = CustomIdentifiedClass()
        self.assertEqual([], obj.foo_int)
        obj.foo_int.append(7)
        obj.foo_int.append(14)
        self.assertEqual([7, 14], obj.foo_int)

        tl_name = 'my_obj'
        tl = CustomTopClass(tl_name)
        tl.children.append(obj)
        doc = sbol3.Document()
        doc.add(tl)
        doc2 = sbol3.Document()
        # Round trip the document
        with tempfile.TemporaryDirectory() as tmpdirname:
            test_file = os.path.join(tmpdirname, 'custom.nt')
            doc.write(test_file, sbol3.NTRIPLES)
            doc2.read(test_file, sbol3.NTRIPLES)
        tl2 = doc2.find(tl_name)
        obj2 = tl2.children[0]
        # The lists are necessarily unordered because of RDF
        # Compare specially
        self.assertCountEqual([7, 14], obj2.foo_int)


if __name__ == '__main__':
    unittest.main()



import posixpath
import unittest

import sbol3


class TestObject(unittest.TestCase):

    def setUp(self) -> None:
        sbol3.set_defaults()

    def test_trailing_slash(self):
        # A trailing slash on an object's identity should automatically be removed
        sbol3.set_namespace('http://example.org/sbol3')
        slash_identity = posixpath.join(sbol3.get_namespace(), 'c1', '')
        self.assertTrue(slash_identity.endswith(posixpath.sep))
        c = sbol3.Component(slash_identity, sbol3.SBO_DNA)
        identity = slash_identity.strip(posixpath.sep)
        self.assertEqual(identity, c.identity)

    def test_copy_properties(self):
        doc = sbol3.Document()
        root = sbol3.Component('root', sbol3.SBO_DNA)
        root.name = 'foo'
        root_copy = root.copy()
        self.assertEqual(root_copy.name, 'foo')

    def test_copy_child_objects(self):
        doc = sbol3.Document()
        root = sbol3.Component('root', sbol3.SBO_DNA)
        sub1 = sbol3.Component('sub1', sbol3.SBO_DNA)
        sub2 = sbol3.Component('sub2', sbol3.SBO_DNA)
        sc1 = sbol3.SubComponent(sub1)
        sc2 = sbol3.SubComponent(sub2)
        root.features.append(sc1)
        root.features.append(sc2)
        doc.add(root)
        doc.add(sub1)
        doc.add(sub2)

        doc2 = sbol3.Document()
        root_copy = root.copy(target_doc=doc2)
        self.assertEqual([sc.identity for sc in root.features],
                         [sc.identity for sc in root_copy.features])


if __name__ == '__main__':
    unittest.main()



import math

from . import *


class VariableFeature(Identified):

    def __init__(self, *,
                 cardinality: str = None,
                 variable: str = None,
                 identity: str = None,
                 type_uri: str = SBOL_VARIABLE_FEATURE) -> None:
        super().__init__(identity, type_uri)
        # Assign default values
        if cardinality is None:
            cardinality = SBOL_ZERO_OR_MORE
        if variable is None:
            variable = PYSBOL3_MISSING
        # Create properties
        self.cardinality = URIProperty(self, SBOL_CARDINALITY, 1, 1,
                                       initial_value=cardinality)
        self.variable = ReferencedObject(self, SBOL_VARIABLE, 1, 1,
                                         initial_value=variable)
        self.variant = ReferencedObject(self, SBOL_VARIANT, 0, math.inf)
        self.variant_collection = ReferencedObject(self, SBOL_VARIANT_COLLECTION,
                                                   0, math.inf)
        self.variant_derivation = ReferencedObject(self, SBOL_VARIANT_DERIVATION,
                                                   0, math.inf)
        self.variant_measure = OwnedObject(self, SBOL_VARIANT_MEASURE,
                                           0, math.inf)
        # Validate
        self.validate()

    def validate(self) -> None:
        super().validate()
        # Cardinality must be in the set of valid URIs
        valid_cardinalities = [SBOL_ONE, SBOL_ONE_OR_MORE, SBOL_ZERO_OR_MORE,
                               SBOL_ZERO_OR_ONE]
        if self.cardinality not in valid_cardinalities:
            raise ValidationError(f'{self.cardinality} is not a valid cardinality')
        if self.variable is None:
            raise ValidationError('VariableComponent.variable is required')



import os
import unittest

import sbol3

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
SBOL3_LOCATION = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL3')


class TestUnitDivision(unittest.TestCase):

    def test_create(self):
        # Note, in practice we would use the already defined
        # http://www.ontology-of-units-of-measure.org/resource/om-2/kilometrePerHour
        display_id = 'udivision'
        symbol = 'k/hr'
        label = 'Kilometers per hour'
        numerator = 'http://www.ontology-of-units-of-measure.org/resource/om-2/kilometre'
        denominator = 'http://www.ontology-of-units-of-measure.org/resource/om-2/hour'
        udiv = sbol3.UnitDivision(display_id, symbol, label, numerator, denominator)
        self.assertIsNotNone(udiv)
        self.assertEqual(symbol, udiv.symbol)
        self.assertEqual(label, udiv.label)
        self.assertEqual(numerator, udiv.numerator)
        self.assertEqual(denominator, udiv.denominator)

    def test_read_from_file(self):
        test_file = os.path.join(SBOL3_LOCATION, 'measurement_entity',
                                 'measurement', 'measurement.nt')
        doc = sbol3.Document()
        doc.read(test_file, sbol3.NTRIPLES)
        uri = 'https://sbolstandard.org/examples/millimolePerLitre'
        udiv = doc.find(uri)
        self.assertIsNotNone(udiv)
        self.assertIsInstance(udiv, sbol3.UnitDivision)
        self.assertEqual('https://sbolstandard.org/examples/millimole', udiv.numerator)
        self.assertEqual('https://sbolstandard.org/examples/litre', udiv.denominator)
        self.assertEqual('millimolar', udiv.name)
        self.assertEqual('millimolar', udiv.label)
        self.assertEqual('millimolePerLitre', udiv.display_id)


class TestUnitExponentiation(unittest.TestCase):

    def test_create(self):
        # Note, in practice we would use the already defined
        # http://www.ontology-of-units-of-measure.org/resource/om-2/cubicMetre
        display_id = 'uexponentiation'
        symbol = 'm3'
        label = 'cubic metre'
        base = 'http://www.ontology-of-units-of-measure.org/resource/om-2/metre'
        exponent = 3
        uexp = sbol3.UnitExponentiation(display_id, symbol, label, base, exponent)
        self.assertIsNotNone(uexp)
        self.assertEqual(symbol, uexp.symbol)
        self.assertEqual(label, uexp.label)
        self.assertEqual(base, uexp.base)
        self.assertEqual(exponent, uexp.exponent)

    def test_read_from_file(self):
        test_file = os.path.join(SBOL3_LOCATION, 'measurement_entity',
                                 'measurement', 'measurement.nt')
        doc = sbol3.Document()
        doc.read(test_file, sbol3.NTRIPLES)
        uri = 'https://sbolstandard.org/examples/cubicMeter'
        uexp = doc.find(uri)
        self.assertIsNotNone(uexp)
        self.assertIsInstance(uexp, sbol3.UnitExponentiation)
        self.assertEqual('https://sbolstandard.org/examples/meter', uexp.base)
        self.assertEqual(3, uexp.exponent)
        self.assertEqual('cubicMeter', uexp.name)
        self.assertEqual('cubicMeter', uexp.label)
        self.assertEqual('cubicMeter', uexp.display_id)


class TestUnitMultiplication(unittest.TestCase):

    def test_create(self):
        # Note, in practice we would use the already defined
        # http://www.ontology-of-units-of-measure.org/resource/om-2/kelvinMole
        display_id = 'umultiplication'
        symbol = 'K mole'
        label = 'kelvin mole'
        term1 = 'https://sbolstandard.org/examples/kelvin'
        term2 = 'https://sbolstandard.org/examples/mole'
        umult = sbol3.UnitMultiplication(display_id, symbol, label, term1, term2)
        self.assertIsNotNone(umult)
        self.assertEqual(symbol, umult.symbol)
        self.assertEqual(label, umult.label)
        self.assertEqual(term1, umult.term1)
        self.assertEqual(term2, umult.term2)

    def test_read_from_file(self):
        test_file = os.path.join(SBOL3_LOCATION, 'measurement_entity',
                                 'measurement', 'measurement.nt')
        doc = sbol3.Document()
        doc.read(test_file, sbol3.NTRIPLES)
        uri = 'https://sbolstandard.org/examples/kelvinMole'
        umult = doc.find(uri)
        self.assertIsNotNone(umult)
        self.assertIsInstance(umult, sbol3.UnitMultiplication)
        self.assertEqual('https://sbolstandard.org/examples/kelvin', umult.term1)
        self.assertEqual('https://sbolstandard.org/examples/mole', umult.term2)
        self.assertEqual('kelvinMole', umult.name)
        self.assertEqual('kelvinMole', umult.label)
        self.assertEqual('kelvinMole', umult.display_id)


if __name__ == '__main__':
    unittest.main()



import math
import posixpath
from typing import Union
from urllib.parse import urlparse

import rdflib

from . import *


class Identified(SBOLObject):

    def __init__(self, identity: str, type_uri: str) -> None:
        super().__init__(identity, type_uri)
        self._display_id = TextProperty(self, SBOL_DISPLAY_ID, 0, 1)
        self.name = TextProperty(self, SBOL_NAME, 0, 1)
        self.description = TextProperty(self, SBOL_DESCRIPTION, 0, 1)
        self.derived_from = URIProperty(self, PROV_DERIVED_FROM, 0, math.inf)
        self.generated_by = URIProperty(self, PROV_GENERATED_BY, 0, math.inf)
        # The type_constraint for measures should really be Measure but
        # that's a circular dependency. Instead we make the type constraint
        # Identified to constrain it somewhat. Identified is the best we
        # can do since every other SBOL class requires Identified to be defined.
        self.measures = OwnedObject(self, SBOL_HAS_MEASURE, 0, math.inf,
                                    type_constraint=Identified)
        # Identity has been set by the SBOLObject constructor
        self._display_id = self._extract_display_id(self.identity)

    @staticmethod
    def _is_valid_display_id(display_id: str) -> bool:
        if display_id is None:
            return True
        # A display id [...] value MUST be composed of only alphanumeric
        # or underscore characters and MUST NOT begin with a digit.
        # (Section 6.1)
        #
        # Make sure everything other than underscores are alphanumeric, and
        # make sure the first character is not a digit.
        #
        # Note: This implies that '_' is not a valid displayId because `isalnum()`
        # will return false if the string is empty, which it will be after we filter
        # out the lone underscore.
        return display_id.replace('_', '').isalnum() and not display_id[0].isdigit()

    @staticmethod
    def _extract_display_id(identity: str) -> Union[None, str]:
        if not identity:
            return None
        parsed = urlparse(identity)
        if not (parsed.scheme and parsed.netloc and parsed.path):
            # if the identity is not a URL, we cannot extract a display id
            # and display id is optional in this case
            return None
        display_id = parsed.path.split('/')[-1]
        if Identified._is_valid_display_id(display_id):
            return display_id
        else:
            msg = f'"{display_id}" is not a valid displayId.'
            msg += ' A displayId MUST be composed of only alphanumeric'
            msg += ' or underscore characters and MUST NOT begin with a digit.'
            raise ValueError(msg)

    def _validate_display_id(self) -> None:
        if self.identity_is_url():
            if (self.display_id is not None and
                    Identified._is_valid_display_id(self.display_id) and
                    self.identity.endswith(self.display_id)):
                return
        else:
            if Identified._is_valid_display_id(self.display_id):
                return
        message = f'{self.display_id} is not a valid displayId for {self.identity}'
        raise ValidationError(message)

    def _update_identity(self, identity: str, display_id: str) -> None:
        """Updates the identity of an Identified when it is added to a
        parent object. SBOL compliant objects and URIs require updating
        whenever an owned object is added to a new parent.
        """
        self._identity = identity
        self._display_id = display_id
        # Now cycle through any owned objects and update their identities
        for _, objects in self._owned_objects.items():
            for child in objects:
                if child.display_id:
                    new_display_id = child.display_id
                else:
                    # Generate a display id based on type and number
                    type_name = child.type_uri[len(SBOL3_NS):]
                    counter_value = self.counter_value(type_name)
                    new_display_id = f'{type_name}{counter_value}'
                new_identity = posixpath.join(self.identity, new_display_id)
                child._update_identity(new_identity, new_display_id)

    def counter_value(self, type_name: str):
        result = 0
        for _, objects in self._owned_objects.items():
            for sibling in objects:
                if sibling.display_id and sibling.display_id.startswith(type_name):
                    counter_string = sibling.display_id[len(type_name):]
                    counter_int = int(counter_string)
                    if counter_int > result:
                        result = counter_int
        return result + 1

    @property
    def display_id(self):
        # display_id is a read only property
        return self._display_id

    def validate(self) -> None:
        super().validate()
        self._validate_display_id()

    def serialize(self, graph: rdflib.Graph):
        identity = rdflib.URIRef(self.identity)
        graph.add((identity, rdflib.RDF.type, rdflib.URIRef(self.type_uri)))
        for prop, items in self._properties.items():
            if not items:
                continue
            rdf_prop = rdflib.URIRef(prop)
            for item in items:
                graph.add((identity, rdf_prop, item))
        for prop, items in self._owned_objects.items():
            if not items:
                continue
            rdf_prop = rdflib.URIRef(prop)
            for item in items:
                graph.add((identity, rdf_prop, rdflib.URIRef(item.identity)))
                item.serialize(graph)



import math

from . import *
# Feature is not exported
from .feature import Feature
from .location import Location
from .typing import *


class SequenceFeature(Feature):

    def __init__(self, locations: List[Location],
                 *, identity: str = None,
                 type_uri: str = SBOL_SEQUENCE_FEATURE) -> None:
        super().__init__(identity, type_uri)
        self.locations: oo_list = OwnedObject(self, SBOL_LOCATION,
                                              1, math.inf,
                                              type_constraint=Location,
                                              initial_value=locations)
        self.validate()

    def validate(self):
        super().validate()
        if len(self.locations) < 1:
            raise ValidationError('LocalSubComponent must have at least 1 location')


def build_sequence_feature(identity: str,
                           *, type_uri: str = SBOL_SEQUENCE_FEATURE) -> SBOLObject:
    obj = SequenceFeature([EntireSequence()], identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[SBOL_LOCATION] = []
    return obj


Document.register_builder(SBOL_SEQUENCE_FEATURE, build_sequence_feature)



import posixpath
import unittest
import sbol3


class TestOwnedObject(unittest.TestCase):

    def test_identity_append(self):
        comp = sbol3.Component('c1', sbol3.SBO_DNA)
        con1_id = 'con1'
        con1 = sbol3.Constraint(sbol3.SBOL_REPLACES,
                                'http://example.com/fake1',
                                'http://example.com/fake2',
                                identity=con1_id)
        expected = posixpath.join(sbol3.get_namespace(), con1_id)
        # The constraint's identity and display_id will be overwritten as
        # part of the append operation. SBOL Compliant URIs (identities)
        # use the class of the object and a counter to generate the
        # display_id.
        expected2 = posixpath.join(comp.identity, 'Constraint1')
        self.assertNotEqual(expected, expected2)
        self.assertEqual(expected, con1.identity)
        # Append should cause the constraint's identity to change
        comp.constraints.append(con1)
        self.assertEqual(expected2, con1.identity)

    def test_list_property_update_identity(self):
        # This test uses assignment instead of appending
        comp = sbol3.Component('c1', sbol3.SBO_DNA)
        con1_id = 'con1'
        con1 = sbol3.Constraint(sbol3.SBOL_REPLACES,
                                'http://example.com/fake1',
                                'http://example.com/fake2',
                                identity=con1_id)
        expected = posixpath.join(sbol3.get_namespace(), con1_id)
        expected2 = posixpath.join(comp.identity, 'Constraint1')
        self.assertNotEqual(expected, expected2)
        self.assertEqual(expected, con1.identity)
        # Setting to list should cause the constraint's identity to change
        comp.constraints = [con1]
        self.assertEqual(expected2, con1.identity)

    def test_singleton_property_update_identity(self):
        tl = sbol3.CustomTopLevel('foo', 'http://synbio.bbn.com/opil#MeasurementValue')
        tl.measure = sbol3.OwnedObject(tl, 'http://synbio.bbn.com/opil#measure', 0, 1)
        m = sbol3.Measure(10, 'liters')
        tl.measure = m
        expected = posixpath.join(tl.identity, 'Measure1')
        self.assertIsNotNone(tl.measure.identity)
        self.assertEqual(expected, tl.measure.identity)

    def test_add_multiple_children(self):
        # Test that the the display_id and identity are overwritten
        # properly when adding multiple child entities.
        comp = sbol3.Component('c1', sbol3.SBO_DNA)
        comp.constraints.append(sbol3.Constraint(sbol3.SBOL_REPLACES,
                                                 'http://example.com/fake1',
                                                 'http://example.com/fake2'))
        expected1 = posixpath.join(comp.identity, 'Constraint1')
        self.assertEqual(expected1, comp.constraints[0].identity)
        comp.constraints.append(sbol3.Constraint(sbol3.SBOL_REPLACES,
                                                 'http://example.com/fake1',
                                                 'http://example.com/fake2'))
        expected2 = posixpath.join(comp.identity, 'Constraint2')
        self.assertEqual(expected2, comp.constraints[1].identity)

    def test_identity_conflict2(self):
        # Test that the same display id will cause a validation
        # error when an item with the same display id is appended
        comp = sbol3.Component('c1', sbol3.SBO_DNA)
        con1_id = 'con1'
        constraints = [sbol3.Constraint(sbol3.SBOL_REPLACES,
                                        'http://example.com/fake1',
                                        'http://example.com/fake2'),
                       sbol3.Constraint(sbol3.SBOL_REPLACES,
                                        'http://example.com/fake1',
                                        'http://example.com/fake2')]
        comp.constraints = constraints
        expected1 = posixpath.join(comp.identity, 'Constraint1')
        self.assertEqual(expected1, comp.constraints[0].identity)
        expected2 = posixpath.join(comp.identity, 'Constraint2')
        self.assertEqual(expected2, comp.constraints[1].identity)

    def test_cascade_identity(self):
        # Test that updating identity of an owned object cascades
        # to child owned objects
        c1 = sbol3.Component('c1', sbol3.SBO_DNA)
        seq = sbol3.Sequence('seq1')
        loc = sbol3.EntireSequence(seq)
        seq_feature = sbol3.SequenceFeature([loc])
        c1.features.append(seq_feature)
        self.assertIsNotNone(seq_feature.identity)
        # identity should cascade down to the location after it
        # is set on the sequence feature
        self.assertIsNotNone(loc.identity)

    def test_type_constraint(self):
        c = sbol3.Component('foo', sbol3.SBO_DNA)
        with self.assertRaises(TypeError):
            c.features = [sbol3.Range('https://example.com/fake', 1, 2)]
        self.assertEqual(c.features.type_constraint, sbol3.Feature)

    # TODO: Write tests for adding via a slice
    #       comp.constraints[0:1] = sbol3.Constraint('foo')


if __name__ == '__main__':
    unittest.main()



import os
import unittest

import sbol3

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
SBOL3_LOCATION = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL3')


class TestExternallyDefined(unittest.TestCase):

    def test_create(self):
        types = ['https://identifiers.org/SBO:0000247']
        definition = 'https://identifiers.org/CHEBI:3312'
        ext_def = sbol3.ExternallyDefined(types, definition)
        self.assertEqual(definition, ext_def.definition)
        self.assertCountEqual(types, ext_def.types)

    def test_read_from_file(self):
        test_file = os.path.join(SBOL3_LOCATION, 'measurement_entity',
                                 'measurement', 'measurement.nt')
        doc = sbol3.Document()
        doc.read(test_file, sbol3.NTRIPLES)
        uri = 'https://sbolstandard.org/examples/M9_Glucose_CAA/CaCl2'
        ext_def = doc.find(uri)
        self.assertIsNotNone(ext_def)
        self.assertIsInstance(ext_def, sbol3.ExternallyDefined)
        self.assertCountEqual(['https://identifiers.org/SBO:0000247'], ext_def.types)
        self.assertEqual('https://identifiers.org/CHEBI:3312', ext_def.definition)
        self.assertEqual('CaCl2', ext_def.display_id)
        self.assertEqual(1, len(ext_def.measures))
        measure = ext_def.measures[0]
        measure_uri = 'https://sbolstandard.org/examples/M9_Glucose_CAA/CaCl2/measure1'
        self.assertEqual(measure_uri, measure.identity)



import math

from . import *


class Collection(TopLevel):

    def __init__(self, identity: str, *, type_uri: str = SBOL_COLLECTION) -> None:
        super().__init__(identity, type_uri)
        self.members = ReferencedObject(self, SBOL_MEMBER, 0, math.inf)


class Namespace(Collection):

    def __init__(self, identity: str, *, type_uri: str = SBOL_NAMESPACE) -> None:
        super().__init__(identity, type_uri=type_uri)


class Experiment(Collection):

    def __init__(self, identity: str, *, type_uri: str = SBOL_EXPERIMENT) -> None:
        super().__init__(identity, type_uri=type_uri)


Document.register_builder(SBOL_COLLECTION, Collection)
Document.register_builder(SBOL_NAMESPACE, Namespace)
Document.register_builder(SBOL_EXPERIMENT, Experiment)



from . import *


class Model(TopLevel):

    def __init__(self, identity: str, *, type_uri: str = SBOL_MODEL) -> None:
        super().__init__(identity, type_uri)
        self.source = URIProperty(self, SBOL_SOURCE, 1, 1)
        self.language = URIProperty(self, SBOL_LANGUAGE, 1, 1)
        self.framework = URIProperty(self, SBOL_FRAMEWORK, 1, 1)

    def validate(self) -> None:
        super().validate()
        # The source property is REQUIRED and MUST contain a URI reference (Section 6.8)
        if not self.source:
            msg = f'Model {self.identity} does not have a source'
            raise ValidationError(msg)
        # The language property is REQUIRED and MUST contain a URI (Section 6.8)
        if not self.language:
            msg = f'Model {self.identity} does not have a language'
            raise ValidationError(msg)
        # The framework property is REQUIRED and MUST contain a URI (Section 6.8)
        if not self.framework:
            msg = f'Model {self.identity} does not have a framework'
            raise ValidationError(msg)


Document.register_builder(SBOL_MODEL, Model)



import unittest

import sbol3


class TestExperimentalData(unittest.TestCase):

    # This class could use some more tests. Not sure what else to put
    # here. There are not yet any examples in the SBOLTestSuite. And
    # the class doesn't have any properties of its own.

    def test_create(self):
        display_id = 'exp_data'
        exp_data = sbol3.ExperimentalData(display_id)
        self.assertIsNotNone(exp_data)
        self.assertEqual(display_id, exp_data.display_id)
        self.assertTrue(hasattr(exp_data, 'attachments'))


if __name__ == '__main__':
    unittest.main()



import unittest

import sbol3
# Reach behind the curtain to test non-public function(s)
from sbol3.utils import parse_class_name


class TestUtils(unittest.TestCase):

    def test_parse_class_name(self):
        # Test parsing with # delimiter in type URI
        c = sbol3.Component('c', sbol3.SBO_DNA)
        class_name = parse_class_name(c.type_uri)
        self.assertEqual(class_name, 'Component')

        # Test parsing with # delimiter in type URI
        m = sbol3.Measure(0, 'dollars')
        class_name = parse_class_name(m.type_uri)
        self.assertEqual(class_name, 'Measure')

        # Test failure with invalid type URI
        with self.assertRaises(ValueError):
            parse_class_name('Component')


if __name__ == '__main__':
    unittest.main()



import posixpath
import unittest
import uuid

import rdflib

import sbol3


class TestIdentified(unittest.TestCase):

    def setUp(self) -> None:
        sbol3.set_defaults()

    def test_display_id(self):
        c1_display_id = 'c1'
        c = sbol3.Component(c1_display_id, sbol3.SBO_DNA)
        self.assertEqual(c1_display_id, c.display_id)
        self.assertIsInstance(c.display_id, str)
        display_id = 'my_component'
        # Ensure that displayId cannot be set
        with self.assertRaises(AttributeError):
            c.display_id = display_id
        # Should still have the old value
        self.assertEqual(c1_display_id, c.display_id)
        # Under the covers
        self.assertIsInstance(c._properties[sbol3.SBOL_DISPLAY_ID][0],
                              rdflib.Literal)

    def test_identity_display_id(self):
        # Test setting of display_id
        #   * Test by passing display_id to constructor
        #   * Test by having display_id deduced from identity
        c1_display_id = 'c1'
        c1_identity = posixpath.join(sbol3.get_namespace(), c1_display_id)
        c1 = sbol3.Component(c1_display_id, sbol3.SBO_DNA)
        self.assertEqual(c1_display_id, c1.display_id)
        self.assertEqual(c1_identity, c1.identity)
        # Now test identity and display_id from a URL-type URI
        c2_display_id = 'c2'
        c2_identity = posixpath.join(sbol3.get_namespace(), c2_display_id)
        c2 = sbol3.Component(c2_identity, sbol3.SBO_DNA)
        self.assertEqual(c2_display_id, c2.display_id)
        self.assertEqual(c2_identity, c2.identity)

    def test_uuid(self):
        # Verify that a UUID can be used as an identity and that
        # the object does not have a display_id since the identity
        # is not a URL
        identity = str(uuid.uuid5(uuid.NAMESPACE_URL, sbol3.get_namespace()))
        c = sbol3.Component(identity, sbol3.SBO_DNA)
        self.assertEqual(identity, c.identity)
        self.assertIsNone(c.display_id)

    def test_basic_serialization(self):
        c = sbol3.Component('c1', sbol3.SBO_DNA)
        graph = rdflib.Graph()
        c.serialize(graph)
        # Is there a better way to get all the triples?
        triples = list(graph.triples((None, None, None)))
        # Expecting a triple for the type, a triple for the displayId,
        # and a triple for the component type
        self.assertEqual(3, len(triples))


if __name__ == '__main__':
    unittest.main()



from . import *


class Implementation(TopLevel):

    def __init__(self, identity: str, *, type_uri: str = SBOL_IMPLEMENTATION) -> None:
        super().__init__(identity, type_uri)
        self.built = ReferencedObject(self, SBOL_BUILT, 0, 1)
        self.validate()

    def validate(self) -> None:
        super().validate()


Document.register_builder(SBOL_IMPLEMENTATION, Implementation)



import os
import unittest

import sbol3

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
SBOL3_LOCATION = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL3')


class TestParticipation(unittest.TestCase):

    def test_create(self):
        participant = 'https://github.com/synbiodex/pysbol3/participant1'
        p = sbol3.Participation([sbol3.SBO_INHIBITOR], participant)
        self.assertIsNotNone(p)
        self.assertEqual([sbol3.SBO_INHIBITOR], p.roles)
        self.assertEqual(participant, p.participant)

    def test_read_from_file(self):
        test_file = os.path.join(SBOL3_LOCATION, 'toggle_switch',
                                 'toggle_switch.nt')
        doc = sbol3.Document()
        doc.read(test_file, sbol3.NTRIPLES)
        search_uri = ('https://sbolstandard.org/examples/'
                      'LacI_producer/interaction_2/participation_1')
        participation = doc.find(search_uri)
        self.assertIsNotNone(participation)
        self.assertIsInstance(participation, sbol3.Participation)
        participant = 'https://sbolstandard.org/examples/LacI_producer/subcomponent_5'
        self.assertEqual(participant, participation.participant)
        roles = [sbol3.SBO_TEMPLATE]
        self.assertEqual(roles, participation.roles)


if __name__ == '__main__':
    unittest.main()



import os
import unittest

import sbol3

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
SBOL3_LOCATION = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL3')


class SingleRefObj(sbol3.TopLevel):

    SRO_URI = 'https://github.com/synbiodex/sbol3#SingleRefObj'
    SEQUENCE_URI = 'https://github.com/synbiodex/sbol3#sequence'

    def __init__(self, identity: str, type_uri: str = SRO_URI):
        super().__init__(identity, type_uri)
        self.sequence = sbol3.ReferencedObject(self, SingleRefObj.SEQUENCE_URI, 0, 1)


class TestReferencedObject(unittest.TestCase):

    def setUp(self) -> None:
        sbol3.set_defaults()

    def test_lookup(self):
        test_path = os.path.join(SBOL3_LOCATION, 'entity', 'model', 'model.ttl')
        test_format = sbol3.TURTLE

        doc = sbol3.Document()
        doc.read(test_path, test_format)
        component = doc.find('toggle_switch')
        self.assertIsNotNone(component)
        model_uri = component.models[0]
        self.assertTrue(isinstance(model_uri, str))
        self.assertTrue(hasattr(model_uri, 'lookup'))
        self.assertEqual('https://sbolstandard.org/examples/model1', model_uri)
        model = model_uri.lookup()
        self.assertIsNotNone(model)

    def test_uri_assignment(self):
        # Test assignment to a ReferencedObject attribute with a URI string
        doc = sbol3.Document()
        component = sbol3.Component('c1', sbol3.SBO_DNA)
        sequence = sbol3.Sequence('seq1')
        doc.add(component)
        doc.add(sequence)
        component.sequences.append(sequence.identity)
        seq2_uri = component.sequences[0]
        self.assertEqual(sequence.identity, seq2_uri)
        seq = seq2_uri.lookup()
        self.assertIsNotNone(seq)
        self.assertEqual(sequence, seq)

    def test_instance_append(self):
        # Test assignment to a ReferencedObject attribute with an
        # instance using append
        doc = sbol3.Document()
        component = sbol3.Component('c1', sbol3.SBO_DNA)
        sequence = sbol3.Sequence('seq1')
        doc.add(component)
        component.sequences.append(sequence)
        seq2_uri = component.sequences[0]
        self.assertEqual(sequence.identity, seq2_uri)
        seq = seq2_uri.lookup()
        self.assertIsNotNone(seq)
        self.assertEqual(sequence, seq)

    def test_instance_assignment(self):
        # Test assignment to a ReferencedObject attribute with an
        # instance using assignment
        doc = sbol3.Document()
        component = sbol3.Component('c1', sbol3.SBO_DNA)
        sequence = sbol3.Sequence('seq1')
        doc.add(component)
        component.sequences = [sequence]
        seq2_uri = component.sequences[0]
        self.assertEqual(sequence.identity, seq2_uri)
        seq = seq2_uri.lookup()
        self.assertIsNotNone(seq)
        self.assertEqual(sequence, seq)

    def test_singleton_assignment(self):
        # Test assignment to a ReferencedObject attribute with an
        # instance using assignment
        doc = sbol3.Document()
        test_parent = SingleRefObj('sro1')
        sequence = sbol3.Sequence('seq1')
        doc.add(test_parent)
        test_parent.sequence = sequence
        seq2_uri = test_parent.sequence
        self.assertEqual(sequence.identity, seq2_uri)
        seq = seq2_uri.lookup()
        self.assertIsNotNone(seq)
        self.assertEqual(sequence, seq)


if __name__ == '__main__':
    unittest.main()



import unittest

import sbol3


class TestSequence(unittest.TestCase):

    def test_create(self):
        display_id = 'seq1'
        seq = sbol3.Sequence(display_id)
        self.assertIsNotNone(seq)
        self.assertEqual(display_id, seq.display_id)
        self.assertIsNone(seq.elements)
        self.assertIsNone(seq.encoding)

    def test_invalid(self):
        display_id = 'seq1'
        seq = sbol3.Sequence(display_id)
        self.assertIsNotNone(seq)
        seq.elements = 'actg'
        with self.assertRaises(sbol3.ValidationError):
            seq.validate()

    def test_valid(self):
        display_id = 'seq1'
        seq = sbol3.Sequence(display_id)
        self.assertIsNotNone(seq)
        seq.elements = 'actg'
        seq.encoding = sbol3.SBOL_IUPAC_DNA
        # Should not raise a ValidationError
        seq.validate()


if __name__ == '__main__':
    unittest.main()



from typing import Union

from . import *
# Feature is not exported
from .feature import Feature


class ComponentReference(Feature):

    def __init__(self, in_child_of: Union[SubComponent, str],
                 feature: Union[Feature, str],
                 *, identity: str = None,
                 type_uri: str = SBOL_COMPONENT_REFERENCE) -> None:
        super().__init__(identity, type_uri)
        self.in_child_of = ReferencedObject(self, SBOL_IN_CHILD_OF, 1, 1,
                                            initial_value=in_child_of)
        self.feature = ReferencedObject(self, SBOL_FEATURES, 1, 1,
                                        initial_value=feature)
        self.validate()

    def validate(self) -> None:
        super().validate()
        # Must have 1 feature
        if self.feature is None:
            raise ValidationError('ComponentReference must have a feature')
        # Must have 1 in_child_of
        if self.in_child_of is None:
            msg = 'ComponentReference must have an in_child_of reference'
            raise ValidationError(msg)


def build_component_reference(identity: str, *,
                              type_uri: str = SBOL_COMPONENT_REFERENCE) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = ComponentReference(missing, missing, identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[SBOL_FEATURES] = []
    obj._properties[SBOL_IN_CHILD_OF] = []
    return obj


Document.register_builder(SBOL_COMPONENT_REFERENCE, build_component_reference)



import math
import unittest

import rdflib

import sbol3


class TestProperty(unittest.TestCase):

    def test_slice_assignment(self):
        c = sbol3.Component('c1', sbol3.SBO_DNA)
        self.assertEqual([], c.roles)
        c.roles.append(sbol3.SO_PROMOTER)
        self.assertEqual([sbol3.SO_PROMOTER], c.roles)
        expected = [rdflib.URIRef(item) for item in c.roles]
        self.assertEqual(expected, c._properties[sbol3.SBOL_ROLE])

        c.roles.append(sbol3.SO_CDS)
        self.assertEqual([sbol3.SO_PROMOTER, sbol3.SO_CDS], c.roles)
        expected = [rdflib.URIRef(item) for item in c.roles]
        self.assertEqual(expected, c._properties[sbol3.SBOL_ROLE])

        c.roles[1] = sbol3.CHEBI_EFFECTOR
        self.assertEqual([sbol3.SO_PROMOTER, sbol3.CHEBI_EFFECTOR], c.roles)
        # Make sure the underlying representation is correct
        expected = [rdflib.URIRef(item) for item in c.roles]
        self.assertEqual(expected, c._properties[sbol3.SBOL_ROLE])

        c.roles.append(sbol3.SO_RBS)
        expected = [sbol3.SO_PROMOTER, sbol3.CHEBI_EFFECTOR, sbol3.SO_RBS]
        self.assertEqual(expected, c.roles)
        expected = [rdflib.URIRef(item) for item in c.roles]
        self.assertEqual(expected, c._properties[sbol3.SBOL_ROLE])

        # Replace the first two elements by slice replacement
        c.roles[0:2] = [sbol3.SO_OPERATOR, sbol3.SO_MRNA]
        expected = [sbol3.SO_OPERATOR, sbol3.SO_MRNA, sbol3.SO_RBS]
        self.assertEqual(expected, c.roles)
        expected = [rdflib.URIRef(item) for item in c.roles]
        self.assertEqual(expected, c._properties[sbol3.SBOL_ROLE])

    def test_boolean_property(self):
        c = sbol3.Component('c1', sbol3.SBO_DNA)
        c.boolean_attribute = sbol3.BooleanProperty(c, 'http://example.org#foo',
                                                    0, 1, [])
        c.boolean_attribute = True
        self.assertEqual(type(c.boolean_attribute), bool)

    def test_bounds(self):
        c = sbol3.Component('c1', sbol3.SBO_DNA)
        c.boolean_attribute = sbol3.BooleanProperty(c, 'http://example.org#foo',
                                                    0, math.inf, [])
        self.assertTrue(hasattr(c.boolean_attribute, '__iter__'))
        with self.assertRaises(TypeError):
            c.boolean_attribute = True
        c.int_attribute = sbol3.IntProperty(c, 'http://example.org#foo',
                                            0, math.inf, [])
        self.assertTrue(hasattr(c.int_attribute, '__iter__'))
        with self.assertRaises(TypeError):
            c.int_attribute = 0


if __name__ == '__main__':
    unittest.main()



import unittest

import sbol3


class TestActivity(unittest.TestCase):

    def test_create(self):
        display_id = 'activity'
        activity = sbol3.Activity(display_id)
        self.assertIsNotNone(activity)
        self.assertEqual(display_id, activity.display_id)
        # attachments come from TopLevel
        self.assertEqual([], activity.attachments)
        self.assertEqual([], activity.types)
        self.assertEqual(None, activity.start_time)
        self.assertEqual(None, activity.end_time)
        self.assertEqual([], activity.usage)
        self.assertEqual([], activity.association)


class TestAgent(unittest.TestCase):

    def test_create(self):
        display_id = 'agent'
        agent = sbol3.Agent(display_id)
        self.assertIsNotNone(agent)
        self.assertEqual(display_id, agent.display_id)
        # attachments come from TopLevel
        self.assertEqual([], agent.attachments)


class TestAssociation(unittest.TestCase):

    def test_create(self):
        agent = sbol3.Agent('agent')
        association = sbol3.Association(agent)
        self.assertIsNotNone(association)
        self.assertEqual([], association.roles)
        self.assertEqual(None, association.plan)
        self.assertEqual(agent.identity, association.agent)


class TestPlan(unittest.TestCase):

    def test_create(self):
        display_id = 'plan'
        plan = sbol3.Plan(display_id)
        self.assertIsNotNone(plan)
        self.assertEqual(display_id, plan.display_id)
        # attachments come from TopLevel
        self.assertEqual([], plan.attachments)


class TestUsage(unittest.TestCase):

    def test_create(self):
        agent = sbol3.Agent('agent')
        usage = sbol3.Usage(agent.identity)
        self.assertIsNotNone(usage)
        self.assertEqual(agent.identity, usage.entity)
        self.assertEqual([], usage.roles)


if __name__ == '__main__':
    unittest.main()



import logging
import os
import tempfile
import unittest

import rdflib

import sbol3

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
SBOL3_LOCATION = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL3')


class TestDocument(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        logging.basicConfig(level=logging.INFO)

    def test_read_ntriples(self):
        # Initial test of Document.read
        filename = 'model.nt'
        test_path = os.path.join(SBOL3_LOCATION, 'entity', 'model',
                                 filename)
        doc = sbol3.Document()
        doc.read(test_path)
        with tempfile.TemporaryDirectory() as tmpdirname:
            test_file = os.path.join(tmpdirname, filename)
            doc.write(test_file, sbol3.NTRIPLES)

    def test_read_turtle(self):
        # Initial test of Document.read
        test_path = os.path.join(SBOL3_LOCATION, 'entity', 'model',
                                 'model.ttl')
        doc = sbol3.Document()
        doc.read(test_path, sbol3.TURTLE)

    def test_read_xml_model(self):
        # Initial test of Document.read
        test_path = os.path.join(SBOL3_LOCATION, 'entity', 'model',
                                 'model.rdf')
        doc = sbol3.Document()
        doc.read(test_path, sbol3.RDF_XML)
        self.assertIsNotNone(doc.find('https://sbolstandard.org/examples/toggle_switch'))
        self.assertIsNotNone(doc.find('toggle_switch'))
        self.assertIsNotNone(doc.find('https://sbolstandard.org/examples/model1'))
        # We have inferred the displayId of model1
        self.assertIsNotNone(doc.find('model1'))

    def test_read_turtle_interface(self):
        # Initial test of Document.read
        test_path = os.path.join(SBOL3_LOCATION, 'entity', 'interface',
                                 'interface.ttl')
        doc = sbol3.Document()
        doc.read(test_path, sbol3.TURTLE)

    def test_read_turtle_toggle_switch(self):
        # Initial test of Document.read
        test_path = os.path.join(SBOL3_LOCATION, 'toggle_switch',
                                 'toggle_switch.ttl')
        doc = sbol3.Document()
        doc.read(test_path, sbol3.TURTLE)

    def test_add(self):
        doc = sbol3.Document()
        type_uri = 'https://github.com/synbiodex/sbol3#TestObj'
        obj1 = sbol3.SBOLObject('obj', type_uri)
        with self.assertRaises(TypeError):
            doc.add(obj1)
        seq = sbol3.Sequence('seq1')
        doc.add(seq)
        seq2 = doc.find(seq.identity)
        self.assertEqual(seq.identity, seq2.identity)

    def test_add_multiple(self):
        # Ensure that duplicate identities cannot be added to the document.
        # See https://github.com/SynBioDex/pySBOL3/issues/39
        document = sbol3.Document()
        namespace1 = sbol3.Namespace(identity=sbol3.SBOL3_NS)
        document.add(namespace1)
        namespace2 = sbol3.Namespace(identity=sbol3.SBOL3_NS)
        with self.assertRaises(ValueError):
            document.add(namespace2)

    def test_write(self):
        doc = sbol3.Document()
        doc.add(sbol3.Component('c1', sbol3.SBO_DNA))
        with tempfile.TemporaryDirectory() as tmpdirname:
            nt_file = os.path.join(tmpdirname, 'test_output.nt')
            doc.write(nt_file, sbol3.NTRIPLES)
            xml_file = os.path.join(tmpdirname, 'test_output.xml')
            doc.write(xml_file, sbol3.RDF_XML)
            # No format specified for these, doc.write now figures
            # that out on its own
            out_path = os.path.join(tmpdirname, 'test_output.rdf')
            doc.write(out_path)
            out_path = os.path.join(tmpdirname, 'test_output.ttl')
            doc.write(out_path)
            out_path = os.path.join(tmpdirname, 'test_output.jsonld')
            doc.write(out_path)

    def test_bind(self):
        doc = sbol3.Document()
        prefix = 'foo'
        ns1 = 'http://example.com/foo'
        doc.bind(prefix, ns1)
        self.assertEqual(ns1, doc._namespaces[prefix])

    def test_add_namespace(self):
        doc = sbol3.Document()
        prefix = 'foo'
        ns1 = 'http://example.com/foo'
        with self.assertWarns(DeprecationWarning):
            doc.addNamespace(ns1, prefix)

    def test_guess_format(self):
        doc = sbol3.Document()
        # Expect ValueError because '.foo' file extension is unknown
        with self.assertRaises(ValueError):
            doc.read('test.foo')
        # Expect ValueError because '.foo' file extension is unknown
        with self.assertRaises(ValueError):
            doc.write('test.foo')

    def test_empty_graph(self):
        # Ensure that an empty document generates an empty RDF graph
        doc = sbol3.Document()
        g = doc.graph()
        self.assertEqual(0, len(g))

    def test_graph(self):
        doc = sbol3.Document()
        doc.add(sbol3.Component('foo', sbol3.SBO_DNA))
        graph = doc.graph()
        self.assertEqual(3, len(graph))
        subjects = set()
        predicates = set()
        for s, p, _ in graph:
            subjects.add(s)
            predicates.add(p)
        # Expecting 1 subject, the component
        self.assertEqual(1, len(subjects))
        # Expecting 3 predicates
        self.assertEqual(3, len(predicates))
        self.assertIn(rdflib.RDF.type, predicates)
        # Convert predicates to strings for the remaining assertions
        predicates = [str(p) for p in predicates]
        self.assertIn(sbol3.SBOL_DISPLAY_ID, predicates)
        self.assertIn(sbol3.SBOL_TYPE, predicates)

    def test_write_string(self):
        # Make sure Document.write_string produces the same output
        # as Document.write. Must use sorted NTriples for this test
        # because other formats do not have a guaranteed order.
        filename = 'model.nt'
        test_path = os.path.join(SBOL3_LOCATION, 'entity', 'model',
                                 filename)
        doc = sbol3.Document()
        doc.read(test_path)
        with tempfile.TemporaryDirectory() as tmpdirname:
            test_file = os.path.join(tmpdirname, filename)
            doc.write(test_file, sbol3.SORTED_NTRIPLES)
            with open(test_file, 'rb') as infile:
                expected = infile.read()
        actual = doc.write_string(sbol3.SORTED_NTRIPLES)
        self.assertEqual(expected, actual)


if __name__ == '__main__':
    unittest.main()



__version__ = '1.0a5'

from .constants import *
from .config import set_defaults, get_namespace, set_namespace
# get_homespace and set_homespace are deprecated and included for backward compatibility
from .config import get_homespace, set_homespace
from .error import *
from .object import SBOLObject
from .property_base import Property, SingletonProperty, ListProperty
from .text_property import TextSingletonProperty, TextProperty
from .int_property import IntProperty
from .float_property import FloatProperty
from .boolean_property import BooleanProperty
from .datetime_property import DateTimeProperty
from .uri_property import URIProperty
from .ownedobject import OwnedObject
from .refobj_property import ReferencedObject
from .identified import Identified
from .toplevel import TopLevel
from .custom import CustomIdentified, CustomTopLevel
from .document import Document
from .constraint import Constraint
from .sequence import Sequence
from .feature import Feature
from .location import Location, Range, Cut, EntireSequence
from .varcomp import VariableFeature
from .participation import Participation
from .interaction import Interaction
from .interface import Interface
from .component import Component
from .model import Model
from .subcomponent import SubComponent
from .localsub import LocalSubComponent
from .seqfeat import SequenceFeature
from .collection import Collection, Namespace, Experiment
from .combderiv import CombinatorialDerivation
from .compref import ComponentReference
from .extdef import ExternallyDefined
from .implementation import Implementation
from .attachment import Attachment
from .expdata import ExperimentalData
from .provenance import Activity, Agent, Association, Plan, Usage
from .om_prefix import BinaryPrefix, SIPrefix
from .om_unit import Measure, PrefixedUnit, SingularUnit
from .om_compound import UnitDivision, UnitExponentiation, UnitMultiplication



import abc
import math

from . import *


class Prefix(CustomTopLevel, abc.ABC):
    """om:Prefix is an abstract base class.

    See Appendix A Section A.2 of the SBOL 3.0 specificiation.
    """

    def __init__(self, identity: str, symbol: str, label: str,
                 factor: float, type_uri: str) -> None:
        super().__init__(identity, type_uri)
        self.symbol = TextProperty(self, OM_SYMBOL, 1, 1,
                                   initial_value=symbol)
        self.alternative_symbols = TextProperty(self, OM_ALTERNATIVE_SYMBOL,
                                                0, math.inf)
        self.label = TextProperty(self, OM_LABEL, 1, 1,
                                  initial_value=label)
        self.alternative_labels = TextProperty(self, OM_ALTERNATIVE_LABEL,
                                               0, math.inf)
        self.comment = TextProperty(self, OM_COMMENT, 0, 1)
        self.long_comment = TextProperty(self, OM_LONG_COMMENT, 0, 1)
        self.factor = FloatProperty(self, OM_HAS_FACTOR, 1, 1,
                                    initial_value=factor)

    def validate(self) -> None:
        super().validate()
        if not self.symbol:
            raise ValidationError('Prefix must contain a symbol')
        if not self.label:
            raise ValidationError('Prefix must contain a label')
        if not self.factor:
            raise ValidationError('Prefix must contain a factor')


class SIPrefix(Prefix):

    def __init__(self, identity: str, symbol: str, label: str,
                 factor: float, *, type_uri: str = OM_SI_PREFIX) -> None:
        super().__init__(identity, symbol, label, factor, type_uri)
        self.validate()


def build_si_prefix(identity: str, *, type_uri: str = OM_SI_PREFIX) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = SIPrefix(identity, missing, missing, 1.0, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[OM_SYMBOL] = []
    obj._properties[OM_LABEL] = []
    obj._properties[OM_HAS_FACTOR] = []
    return obj


Document.register_builder(OM_SI_PREFIX, build_si_prefix)


class BinaryPrefix(Prefix):

    def __init__(self, identity: str, symbol: str, label: str,
                 factor: float, *, type_uri: str = OM_BINARY_PREFIX) -> None:
        super().__init__(identity, symbol, label, factor, type_uri)
        self.validate()


def build_binary_prefix(identity: str,
                        *, type_uri: str = OM_BINARY_PREFIX) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = BinaryPrefix(identity, missing, missing, 1.0, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[OM_SYMBOL] = []
    obj._properties[OM_LABEL] = []
    obj._properties[OM_HAS_FACTOR] = []
    return obj


Document.register_builder(OM_BINARY_PREFIX, build_binary_prefix)



import rdflib

from . import *


class CustomIdentified(Identified):

    def __init__(self, type_uri: str = None, *, identity: str = None,
                 sbol_type_uri: str = SBOL_IDENTIFIED) -> None:
        super().__init__(identity, type_uri)
        self.rdf_type = URIProperty(self, rdflib.RDF.type, 1, 1,
                                    initial_value=sbol_type_uri)

    def validate(self) -> None:
        super().validate()
        if self.rdf_type is None:
            raise ValidationError('rdf_type is a required property of CustomIdentified')


class CustomTopLevel(TopLevel):

    def __init__(self, identity: str = None, type_uri: str = None,
                 *, sbol_type_uri: str = SBOL_TOP_LEVEL) -> None:
        super().__init__(identity, type_uri)
        self.rdf_type = URIProperty(self, rdflib.RDF.type, 1, 1,
                                    initial_value=sbol_type_uri)

    def validate(self) -> None:
        super().validate()
        if self.rdf_type is None:
            raise ValidationError('rdf_type is a required property of CustomTopLevel')



import abc
import collections
from collections import MutableSequence, Iterable
from typing import Any, Optional, List, Dict, Union


class Property(abc.ABC):

    def __init__(self, property_owner: Any, property_uri: str,
                 lower_bound: int, upper_bound: int,
                 validation_rules: Optional[List] = None) -> None:
        self.property_owner = property_owner
        self.property_uri = property_uri
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        if validation_rules is None:
            validation_rules = []
        self.validation_rules = validation_rules
        # Initialize the storage for this property
        # Do not overwrite a value if already present (see issue #76)
        if self.property_uri not in self._storage():
            self._storage()[self.property_uri] = []

    def _storage(self) -> Dict[str, list]:
        return self.property_owner._properties

    @abc.abstractmethod
    def set(self, value: Any) -> None:
        pass

    @abc.abstractmethod
    def to_user(self, value: Any) -> str:
        pass

    @abc.abstractmethod
    def from_user(self, value: Any):
        pass

    def item_added(self, item: Any) -> None:
        """Stub method for child classes to override if they have to do
        any additional processing on items after they are added. This method
        will be called on each individual item that was added to the list.
        """
        pass


class SingletonProperty(Property, abc.ABC):

    def __init__(self, property_owner: Any, property_uri: str,
                 lower_bound: int, upper_bound: int,
                 validation_rules: Optional[List] = None) -> None:
        super().__init__(property_owner, property_uri,
                         lower_bound, upper_bound,
                         validation_rules)
        # This is a marker for SBOLObject.__get_attribute__
        self._sbol_singleton = True

    def set(self, value: Any) -> None:
        value = self.from_user(value)
        if value is None:
            if self.lower_bound == 0:
                self._storage()[self.property_uri] = []
            else:
                raise ValueError(f'Property {self.property_uri} cannot be unset')
        else:
            self._storage()[self.property_uri] = [value]
            self.item_added(value)

    def get(self) -> Any:
        try:
            value = self._storage()[self.property_uri][0]
        except IndexError:
            return None
        return self.to_user(value)


class ListProperty(Property, MutableSequence, abc.ABC):

    def __delitem__(self, key: Union[int, slice]) -> None:
        self._storage()[self.property_uri].__delitem__(key)

    def __setitem__(self, key: Union[int, slice], value: Any) -> None:
        # Do string separately because it, too, is iterable
        if isinstance(value, str):
            values = [value]
            value = self.from_user(value)
        elif isinstance(value, Iterable):
            # string is iterable so it's broken out above
            values = value
            value = [self.from_user(item) for item in value]
        else:
            # Not string or iterable
            values = [value]
            value = self.from_user(value)
        self._storage()[self.property_uri].__setitem__(key, value)
        for val in values:
            self.item_added(val)

    def __getitem__(self, key: Union[int, slice]) -> Any:
        value = self._storage()[self.property_uri].__getitem__(key)
        if isinstance(value, str):
            return self.to_user(value)
        elif isinstance(value, collections.Iterable):
            return [self.to_user(v) for v in value]
        else:
            # Not a string or an iterable, just convert
            return self.to_user(value)

    def __len__(self) -> int:
        return self._storage()[self.property_uri].__len__()

    def __contains__(self, item) -> bool:
        item = self.from_user(item)
        return self._storage()[self.property_uri].__contains__(item)

    def __eq__(self, other) -> bool:
        storage = self._storage()[self.property_uri]
        value = [self.to_user(v) for v in storage]
        return value.__eq__(other)

    def __str__(self) -> str:
        return str([self.to_user(item)
                    for item in self._storage()[self.property_uri]])

    def __repr__(self) -> str:
        return repr([self.to_user(item)
                     for item in self._storage()[self.property_uri]])

    def insert(self, index: int, value: Any) -> None:
        item = self.from_user(value)
        self._storage()[self.property_uri].insert(index, item)
        self.item_added(value)

    def set(self, value: Any) -> None:
        # TODO: validate here
        # TODO: test for iterable or sequence types, then convert to list?
        items = [self.from_user(v) for v in value]
        self._storage()[self.property_uri] = items
        for val in value:
            self.item_added(val)



import warnings
from urllib.parse import urlparse

# We probably want to turn this into a dict like the configuration dict
# in pySBOL2. When we have more variables to store, let's change this
# to a dict.
SBOL3_NAMESPACE = ''


def get_namespace() -> str:
    return SBOL3_NAMESPACE


def set_namespace(namespace: str) -> None:
    parsed = urlparse(namespace)
    if parsed.scheme and parsed.netloc:
        global SBOL3_NAMESPACE
        SBOL3_NAMESPACE = namespace
    else:
        raise ValueError(f'Expected URL found {namespace}')


def set_defaults() -> None:
    """Restores configuration to factory settings.
    """
    set_namespace('http://example.com/sbol3/')


set_defaults()


def set_homespace(homespace: str) -> None:
    warnings.warn('Use set_namespace instead', DeprecationWarning)
    return set_namespace(homespace)


def get_homespace() -> str:
    warnings.warn('Use get_namespace instead', DeprecationWarning)
    return get_namespace()



import math
from typing import Union

from . import *


class CombinatorialDerivation(TopLevel):

    def __init__(self, identity: str, template: Union[Component, str],
                 *, type_uri: str = SBOL_MODEL) -> None:
        super().__init__(identity, type_uri)
        self.strategy = URIProperty(self, SBOL_STRATEGY, 0, 1)
        self.template = ReferencedObject(self, SBOL_TEMPLATE, 1, 1,
                                         initial_value=template)
        self.variable_components = OwnedObject(self, SBOL_VARIABLE_FEATURES,
                                               0, math.inf,
                                               type_constraint=VariableFeature)
        self.validate()

    def validate(self) -> None:
        super().validate()
        if self.strategy is not None:
            valid_strategies = [SBOL_ENUMERATE, SBOL_SAMPLE]
            if self.strategy not in valid_strategies:
                raise ValidationError(f'{self.strategy} is not a valid strategy')


def build_combinatorial_derivation(identity: str,
                                   *, type_uri: str = SBOL_COMBINATORIAL_DERIVATION):
    template = PYSBOL3_MISSING
    return CombinatorialDerivation(identity, template, type_uri=type_uri)


Document.register_builder(SBOL_COMBINATORIAL_DERIVATION, build_combinatorial_derivation)



from typing import Optional, Union

from . import *


class Constraint(Identified):

    def __init__(self, restriction: str, subject: Union[Identified, str],
                 object: Union[Identified, str], *, identity: Optional[str] = None,
                 type_uri: str = SBOL_CONSTRAINT) -> None:
        super().__init__(identity, type_uri)
        self.restriction = URIProperty(self, SBOL_RESTRICTION, 1, 1,
                                       initial_value=restriction)
        self.subject = ReferencedObject(self, SBOL_SUBJECT, 1, 1,
                                        initial_value=subject)
        self.object = ReferencedObject(self, SBOL_OBJECT, 1, 1,
                                       initial_value=object)
        self.validate()

    def validate(self) -> None:
        super().validate()
        if not self.restriction:
            raise ValidationError('Constraint must have a restriction')
        if not self.subject:
            raise ValidationError('Constraint must have a subject')
        if not self.object:
            raise ValidationError('Constraint must have an object')


def build_constraint(identity: str, type_uri: str = SBOL_CONSTRAINT) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = Constraint(missing, missing, missing, identity=identity, type_uri=type_uri)
    obj._properties[SBOL_RESTRICTION] = []
    obj._properties[SBOL_SUBJECT] = []
    obj._properties[SBOL_OBJECT] = []
    return obj


Document.register_builder(SBOL_CONSTRAINT, build_constraint)



import unittest

import rdflib

import sbol3


class TestCollection(unittest.TestCase):

    def test_create(self):
        collection = sbol3.Collection('collection1')
        self.assertIsNotNone(collection)
        self.assertEqual(0, len(collection.members))

    def test_read(self):
        identity = 'https://github.com/synbiodex/pysbol3/collection1'
        nt_data = f'<{identity}> <{rdflib.RDF.type}> <{sbol3.SBOL_COLLECTION}> .'
        doc = sbol3.Document()
        doc.read_string(nt_data, 'ttl')
        collection = doc.find(identity)
        self.assertIsNotNone(collection)
        self.assertIsInstance(collection, sbol3.Collection)

    def test_member_property(self):
        self.assertTrue(hasattr(sbol3, 'SBOL_MEMBER'))
        collection = sbol3.Collection('collection1')
        self.assertIn(sbol3.SBOL_MEMBER, collection._properties)
        self.assertNotIn(sbol3.SBOL_ORIENTATION, collection._properties)
        uris = ['https://github.com/synbiodex/pysbol3/thing1',
                'https://github.com/synbiodex/pysbol3/thing2']
        collection.members = uris
        self.assertIn(sbol3.SBOL_MEMBER, collection._properties)
        self.assertNotIn(sbol3.SBOL_ORIENTATION, collection._properties)
        self.assertEqual(uris, collection.members)


class TestNamespace(unittest.TestCase):

    def test_create(self):
        namespace = sbol3.Namespace('namespace1')
        self.assertIsNotNone(namespace)
        self.assertEqual(0, len(namespace.members))

    def test_read(self):
        identity = 'https://github.com/synbiodex/pysbol3/namespace1'
        nt_data = f'<{identity}> <{rdflib.RDF.type}> <{sbol3.SBOL_NAMESPACE}> .'
        doc = sbol3.Document()
        doc.read_string(nt_data, 'ttl')
        namespace = doc.find(identity)
        self.assertIsNotNone(namespace)
        self.assertIsInstance(namespace, sbol3.Namespace)


class TestExperiment(unittest.TestCase):

    def test_create(self):
        experiment = sbol3.Experiment('experiment1')
        self.assertIsNotNone(experiment)
        self.assertEqual(0, len(experiment.members))

    def test_read(self):
        identity = 'https://github.com/synbiodex/pysbol3/experiment1'
        nt_data = f'<{identity}> <{rdflib.RDF.type}> <{sbol3.SBOL_EXPERIMENT}> .'
        doc = sbol3.Document()
        doc.read_string(nt_data, 'ttl')
        experiment = doc.find(identity)
        self.assertIsNotNone(experiment)
        self.assertIsInstance(experiment, sbol3.Collection)


if __name__ == '__main__':
    unittest.main()



import unittest
from collections import Container

import sbol3


class TestComponent(unittest.TestCase):

    def test_roles(self):
        c = sbol3.Component('c1', sbol3.SBO_DNA)
        self.assertListEqual([], list(c.roles))
        self.assertEqual([], c.roles)
        c.roles = [sbol3.SO_PROMOTER]
        # Make sure the underlying datatype did not get overwritten
        self.assertNotIsInstance(c.roles, list)
        self.assertIsInstance(c.roles, Container)
        self.assertIn(sbol3.SO_PROMOTER, c.roles)
        self.assertEqual([sbol3.SO_PROMOTER], c.roles)
        self.assertTrue(c.roles == [sbol3.SO_PROMOTER])
        self.assertTrue([sbol3.SO_PROMOTER] == c.roles)
        # Other list manipulations
        c.roles.append(sbol3.SO_CDS)
        self.assertEqual([sbol3.SO_PROMOTER, sbol3.SO_CDS], c.roles)
        self.assertEqual([sbol3.SO_CDS], c.roles[1:])
        c.roles[1:] = [sbol3.SO_RBS]
        self.assertEqual([sbol3.SO_PROMOTER, sbol3.SO_RBS], c.roles)
        self.assertEqual([sbol3.SO_RBS], c.roles[1:])


if __name__ == '__main__':
    unittest.main()



from . import *


class Sequence(TopLevel):

    def __init__(self, identity: str, *, type_uri: str = SBOL_SEQUENCE) -> None:
        super().__init__(identity, type_uri)
        self.elements = TextProperty(self, SBOL_ELEMENTS, 0, 1)
        self.encoding = URIProperty(self, SBOL_ENCODING, 0, 1)
        self.validate()

    def validate(self) -> None:
        super().validate()
        # If sequence is set, encoding is REQUIRED
        if self.elements and not self.encoding:
            raise ValidationError('Sequence encoding is required if elements are set')


Document.register_builder(SBOL_SEQUENCE, Sequence)



import math

from . import *


class Interface(Identified):

    def __init__(self, *, identity: str = None,
                 type_uri: str = SBOL_INTERFACE) -> None:
        super().__init__(identity, type_uri)
        self.input = ReferencedObject(self, SBOL_INPUT, 0, math.inf)
        self.output = ReferencedObject(self, SBOL_OUTPUT, 0, math.inf)
        self.non_directional = ReferencedObject(self, SBOL_NON_DIRECTIONAL, 0, math.inf)
        self.validate()

    def validate(self) -> None:
        super().validate()


def build_interface(identity: str, *, type_uri: str = SBOL_INTERFACE) -> SBOLObject:
    return Interface(identity=identity, type_uri=type_uri)


Document.register_builder(SBOL_INTERFACE, build_interface)



import unittest

import sbol3


class TestConfig(unittest.TestCase):

    # I wanted to have this in a base class but I couldn't get the
    # imports right. I kept getting "attempted relative import with
    # no known parent package".
    def setUp(self) -> None:
        sbol3.set_defaults()

    def test_namespace(self):
        base_uri = 'https://github.com/synbiodex/pysbol3'
        sbol3.set_namespace(base_uri)
        self.assertEqual(base_uri, sbol3.get_namespace())

        # Example from SBOL 3.0 Section 5.1 (page 12)
        # See issue #80
        base_uri = 'https://synbiohub.org'
        sbol3.set_namespace(base_uri)
        self.assertEqual(base_uri, sbol3.get_namespace())


if __name__ == '__main__':
    unittest.main()



import unittest

import sbol3

OM_KILOGRAM = 'http://www.ontology-of-units-of-measure.org/resource/om-2/kilogram'


class TestModule(unittest.TestCase):

    def test_data_model(self):
        self.assertIsInstance(sbol3.Identified, type)
        self.assertIsInstance(sbol3.TopLevel, type)
        self.assertIsInstance(sbol3.Sequence, type)
        self.assertIsInstance(sbol3.Component, type)
        self.assertIsInstance(sbol3.SubComponent, type)
        self.assertIsInstance(sbol3.ComponentReference, type)
        self.assertIsInstance(sbol3.LocalSubComponent, type)
        self.assertIsInstance(sbol3.ExternallyDefined, type)
        self.assertIsInstance(sbol3.SequenceFeature, type)
        self.assertIsInstance(sbol3.Range, type)
        self.assertIsInstance(sbol3.Cut, type)
        self.assertIsInstance(sbol3.EntireSequence, type)
        self.assertIsInstance(sbol3.Constraint, type)
        self.assertIsInstance(sbol3.Interaction, type)
        self.assertIsInstance(sbol3.Participation, type)
        self.assertIsInstance(sbol3.Interface, type)
        self.assertIsInstance(sbol3.CombinatorialDerivation, type)
        self.assertIsInstance(sbol3.VariableFeature, type)
        self.assertIsInstance(sbol3.Implementation, type)
        self.assertIsInstance(sbol3.ExperimentalData, type)
        self.assertIsInstance(sbol3.Model, type)
        self.assertIsInstance(sbol3.Collection, type)
        self.assertIsInstance(sbol3.Namespace, type)
        self.assertIsInstance(sbol3.Experiment, type)
        self.assertIsInstance(sbol3.Attachment, type)

    def test_provenance(self):
        self.assertIsInstance(sbol3.Activity, type)
        self.assertIsInstance(sbol3.Association, type)
        self.assertIsInstance(sbol3.Plan, type)
        self.assertIsInstance(sbol3.Agent, type)
        self.assertIsInstance(sbol3.Usage, type)

    def test_measurement(self):
        self.assertIsInstance(sbol3.Measure, type)
        self.assertIsInstance(sbol3.SingularUnit, type)
        self.assertIsInstance(sbol3.UnitMultiplication, type)
        self.assertIsInstance(sbol3.UnitDivision, type)
        self.assertIsInstance(sbol3.UnitExponentiation, type)
        self.assertIsInstance(sbol3.PrefixedUnit, type)
        self.assertIsInstance(sbol3.SIPrefix, type)
        self.assertIsInstance(sbol3.BinaryPrefix, type)

    def test_identified_constructors(self):
        # I don't remember why I wanted this test.
        # Test all of the Identified constructors.
        args_map = {
            'Association': ['https://example.com/fake'],
            'ComponentReference': ['https://example.com/fake',
                                   'https://example.com/fake'],
            'Constraint': ['https://example.com/restriction',
                           'https://example.com/subject',
                           'https://example.com/object'],
            'CustomIdentified': ['https://example.com/fake'],
            'Cut': ['https://example.com/fake',
                    1],
            'EntireSequence': ['https://example.com/fake'],
            'ExternallyDefined': [['https://example.com/fake'],
                                  'https://example.com/fake'],
            'Interaction': [['https://example.com/fake']],
            'LocalSubComponent': [['https://example.com/fake']],
            'Measure': [1.0, OM_KILOGRAM],
            'Participation': [sbol3.SBO_INHIBITOR,
                              'https://example.com/fake'],
            'Range': ['https://example.com/fake', 1, 2],
            'SequenceFeature': [[sbol3.EntireSequence('https://example.com/fake')]],
            'SubComponent': ['https://example.com/fake'],
            'Usage': ['https://example.com/fake'],
        }
        skip_list = [sbol3.Identified, sbol3.Feature, sbol3.Location]
        for name in dir(sbol3):
            item = getattr(sbol3, name)
            if not isinstance(item, type):
                continue
            if issubclass(item, sbol3.TopLevel):
                continue
            if not issubclass(item, sbol3.Identified):
                continue
            if item in skip_list:
                continue
            arg_list = []
            if name in args_map:
                arg_list = args_map[name]
            try:
                item(*arg_list)
            except TypeError as e:
                self.fail('Unable to construct sbol3.%s: %s' % (name, e))
            except sbol3.ValidationError as e:
                self.fail('Constructed invalid sbol3.%s: %s' % (name, e))


if __name__ == '__main__':
    unittest.main()



import collections
import logging
import warnings
from typing import Dict, Callable, List, Optional

import rdflib
# Get the rdflib-jsonld capability initialized
# Note: this is for side effect. The parser is not used.
# The side effect is that the JSON-LD parser is registered in RDFlib.
from rdflib_jsonld import parser as jsonld_parser

from . import *
from .object import BUILDER_REGISTER

_default_bindings = {
    'sbol': SBOL3_NS,
    'prov': PROV_NS,
    'om': OM_NS,
    # Should others like SO, SBO, and CHEBI be added?
}


class Document:

    @staticmethod
    def register_builder(type_uri: str,
                         builder: Callable[[str, str], SBOLObject]) -> None:
        """A builder function will be called with an identity and a
        keyword argument type_uri.

        builder(identity_uri: str, type_uri: str = None) -> SBOLObject
        """
        Document._uri_type_map[type_uri] = builder

    # Map type URIs to a builder function to construct entities from
    # RDF triples.
    _uri_type_map: Dict[str, Callable[[str, str], SBOLObject]] = BUILDER_REGISTER

    def __init__(self):
        self.logger = logging.getLogger(SBOL_LOGGER_NAME)
        self.objects: List[Identified] = []
        self._namespaces: Dict[str, str] = _default_bindings.copy()

    def _make_custom_object(self, identity: str, types: List[str]) -> Identified:
        if SBOL_IDENTIFIED in types:
            types.remove(SBOL_IDENTIFIED)
            try:
                other_type = types[0]
            except IndexError:
                raise ValidationError('Expected one other type')
            if other_type.startswith(SBOL3_NS):
                raise ValidationError('Secondary type may not be in SBOL3 namespace')
            try:
                builder = self._uri_type_map[other_type]
            except KeyError:
                logging.warning(f'No builder found for {other_type}')
                builder = CustomIdentified
            return builder(identity=identity, type_uri=other_type)
        elif SBOL_TOP_LEVEL in types:
            types.remove(SBOL_TOP_LEVEL)
            try:
                other_type = types[0]
            except IndexError:
                raise ValidationError('Expected one other type')
            if other_type.startswith(SBOL3_NS):
                raise ValidationError('Secondary type may not be in SBOL3 namespace')
            try:
                builder = self._uri_type_map[other_type]
            except KeyError:
                logging.warning(f'No builder found for {other_type}')
                builder = CustomTopLevel
            return builder(identity=identity, type_uri=other_type)
        else:
            message = 'Custom types must contain either Identified or TopLevel'
            raise ValidationError(message)

    def _parse_objects(self, graph: rdflib.Graph) -> Dict[str, SBOLObject]:
        # First extract the identities and their types. Each identity
        # can have either one or two rdf:type properties. If one,
        # create the entity. If two, it is a custom type (see section
        # 6.11 of the spec) and we instantiate it specially.
        #
        # This will have to change in the future when we enable
        # user-defined custom types somehow.
        identity_types: Dict[str, List[str]] = collections.defaultdict(list)
        for s, p, o in graph.triples((None, rdflib.RDF.type, None)):
            str_o = str(o)
            str_s = str(s)
            identity_types[str_s].append(str_o)
        # Now iterate over the identity->type dict creating the objects.
        result = {}
        for identity, types in identity_types.items():
            if len(types) == 1:
                type_uri = types[0]
                try:
                    builder = self._uri_type_map[type_uri]
                except KeyError:
                    logging.warning(f'No builder found for {type_uri}')
                    raise ValidationError(f'Unknown type {type_uri}')
                obj = builder(identity=identity, type_uri=type_uri)
            elif len(types) == 2:
                obj = self._make_custom_object(identity, types)
            else:
                message = f'Expected either 1 or 2 types for {identity}'
                raise ValidationError(message)
            obj.document = self
            result[obj.identity] = obj
        return result

    @staticmethod
    def _parse_attributes(objects, graph):
        for s, p, o in graph.triples((None, None, None)):
            if p == rdflib.RDF.type:
                # RDF.types have been processed already
                continue
            str_s = str(s)
            str_p = str(p)
            obj = objects[str_s]
            if str_p in obj._owned_objects:
                other_identity = str(o)
                other = objects[other_identity]
                obj._owned_objects[str_p].append(other)
            else:
                obj._properties[str_p].append(o)

    @staticmethod
    def _clean_up_singletons(objects: Dict[str, SBOLObject]):
        """Clean up singleton properties after reading an SBOL file.

        When an SBOL file is read, values are appended to the property
        stores without knowledge of which stores are singletons and
        which stores are lists. This method cleans up singleton properties
        by ensuring that each has exactly one value.
        """
        # This is necessary due to defaulting of properties when using
        # the builder. Some objects have required properties, which the
        # builder sets. In the case of singleton values, that can result
        # in multiple values in a singleton property. Only the first value
        # is used, so the value read from file is ignored.
        for _, obj in objects.items():
            for name, attr in obj.__dict__.items():
                if isinstance(attr, SingletonProperty):
                    prop_uri = attr.property_uri
                    store = attr._storage()
                    if len(store[prop_uri]) > 1:
                        store[prop_uri] = store[prop_uri][-1:]

    def clear(self) -> None:
        self.objects.clear()
        self._namespaces = _default_bindings.copy()

    def _parse_graph(self, graph) -> None:
        objects = self._parse_objects(graph)
        self._parse_attributes(objects, graph)
        self._clean_up_singletons(objects)
        # Validate all the objects
        for obj in objects.values():
            obj.validate()
        # Store the TopLevel objects in the Document
        self.objects = [obj for uri, obj in objects.items()
                        if isinstance(obj, TopLevel)]
        # Store the namespaces in the Document for later use
        for prefix, uri in graph.namespaces():
            self.bind(prefix, uri)

    def _guess_format(self, fpath: str):
        return rdflib.util.guess_format(fpath)

    # Formats: 'n3', 'nt', 'turtle', 'xml'
    def read(self, location: str, file_format: str = None) -> None:
        if file_format is None:
            file_format = self._guess_format(location)
        if file_format is None:
            raise ValueError('Unable to determine file format')
        if file_format == SORTED_NTRIPLES:
            file_format = NTRIPLES
        graph = rdflib.Graph()
        graph.load(location, format=file_format)
        return self._parse_graph(graph)

    # Formats: 'n3', 'nt', 'turtle', 'xml'
    def read_string(self, data: str, file_format: str) -> None:
        # TODO: clear the document, this isn't append
        if file_format == SORTED_NTRIPLES:
            file_format = NTRIPLES
        graph = rdflib.Graph()
        graph.parse(data=data, format=file_format)
        return self._parse_graph(graph)

    def add(self, obj: TopLevel) -> None:
        """Add objects to the document.
        """
        if not isinstance(obj, TopLevel):
            message = f'Expected TopLevel instance, {type(obj).__name__} found'
            raise TypeError(message)
        found_obj = self.find(obj.identity)
        if found_obj is not None:
            message = f'An entity with identity "{obj.identity}"'
            message += ' already exists in document'
            raise ValueError(message)
        self.objects.append(obj)
        obj.document = self

    def _find_in_objects(self, search_string: str) -> Optional[Identified]:
        # TODO: implement recursive search
        for obj in self.objects:
            # TODO: needs an object.find(search_string) method on ... Identified?
            result = obj.find(search_string)
            if result is not None:
                return result
        return None

    def find(self, search_string: str) -> Optional[Identified]:
        """Find an object by identity URI or by display_id.

        :param search_string: Either an identity URI or a display_id
        :type search_string: str
        :returns: The named object or ``None`` if no object was found

        """
        for obj in self.objects:
            if obj.identity == search_string:
                return obj
            if obj.display_id and obj.display_id == search_string:
                return obj
        return self._find_in_objects(search_string)

    def write_string(self, file_format: str) -> str:
        graph = self.graph()
        if file_format == SORTED_NTRIPLES:
            # have RDFlib give us the ntriples as a string
            nt_text = graph.serialize(format='nt')
            # split it into lines
            lines = nt_text.splitlines(keepends=True)
            # sort those lines
            lines.sort()
            # write out the lines
            # RDFlib gives us bytes, so open file in binary mode
            result = b''.join(lines)
        elif file_format == JSONLD:
            context = {f'@{prefix}': uri for prefix, uri in self._namespaces.items()}
            context['@vocab'] = 'https://sbolstandard.org/examples/'
            result = graph.serialize(format=file_format, context=context)
        else:
            result = graph.serialize(format=file_format)
        return result

    def write(self, fpath: str, file_format: str = None) -> None:
        """Write the document to file.

        If file_format is None the desired format is guessed from the
        extension of fpath. If file_format cannot be guessed a ValueError
        is raised.
        """
        if file_format is None:
            file_format = self._guess_format(fpath)
        if file_format is None:
            raise ValueError('Unable to determine file format')
        with open(fpath, 'wb') as outfile:
            outfile.write(self.write_string(file_format))

    def graph(self) -> rdflib.Graph:
        """Convert document to an RDF Graph.

        The returned graph is a snapshot of the document and will
        not be updated by subsequent changes to the document.
        """
        graph = rdflib.Graph()
        for prefix, uri in self._namespaces.items():
            graph.bind(prefix, uri)
        for obj in self.objects:
            obj.serialize(graph)
        return graph

    def bind(self, prefix: str, uri: str) -> None:
        """Bind a prefix to an RDF namespace in the written RDF document.

        These prefixes make the written RDF easier for humans to read.
        These prefixes do not change the semantic meaning of the RDF
        document in any way.
        """
        # Remove any prefix referencing the given URI
        if uri in self._namespaces.values():
            for k, v in list(self._namespaces.items()):
                if v == uri:
                    del self._namespaces[k]
        self._namespaces[prefix] = uri

    def addNamespace(self, namespace: str, prefix: str) -> None:
        """Document.addNamespace is deprecated. Replace with Document.bind.

        Document.addNamespace existed in pySBOL2 and was commonly used.

        Document.addNamespace(namespace, prefix) should now be
        Document.bind(prefix, namespace). Note the change of argument
        order.
        """
        warnings.warn('Use Document.bind() instead', DeprecationWarning)
        self.bind(prefix, namespace)



import logging
import os
import shutil
import tempfile
import unittest

import rdflib
import rdflib.compare

import sbol3

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
SBOL_TEST_SUITE = os.path.join(MODULE_LOCATION, 'SBOLTestSuite')
SBOL3_LOCATION = os.path.join(SBOL_TEST_SUITE, 'SBOL3')

DEBUG_ENV_VAR = 'SBOL_TEST_DEBUG'


class TestRoundTrip(unittest.TestCase):

    def setUp(self):
        # Create temp directory
        self.temp_out_dir = tempfile.mkdtemp()
        self.logger = logging.getLogger('sbol3.test')
        if not self.logger.hasHandlers():
            logging.basicConfig()
        if DEBUG_ENV_VAR in os.environ:
            self.logger.setLevel(logging.DEBUG)

    def tearDown(self):
        # Remove directory after the test
        if self.temp_out_dir:
            shutil.rmtree(self.temp_out_dir)
            self.temp_out_dir = None

    def test_BBa_F2620_PoPSReceiver(self):
        sbol_path = os.path.join(SBOL3_LOCATION, 'BBa_F2620_PoPSReceiver',
                                 'BBa_F2620_PoPSReceiver.ttl')
        doc = sbol3.Document()
        doc.read(sbol_path, sbol3.TURTLE)
        uri = 'https://synbiohub.org/public/igem/BBa_F2620/subcomponent_3/location_1'
        range1 = doc.find(uri)
        self.assertIsNotNone(range1)
        self.assertEqual(55, range1.start)
        self.assertEqual(108, range1.end)

    def find_all_files(self, dirname: str):
        for item in os.listdir(dirname):
            item_path = os.path.join(dirname, item)
            if os.path.isdir(item_path):
                for f in self.find_all_files(item_path):
                    yield f
            elif os.path.isfile(item_path):
                yield item_path
            else:
                print(f'{item} is neither file nor directory')

    @staticmethod
    def rdf_type(filename: str):
        filename = os.path.basename(filename)
        ext = os.path.splitext(filename)[1]
        # drop the leading dot
        ext = ext[1:]
        ext_map = {'nt': sbol3.NTRIPLES,
                   'ttl': sbol3.TURTLE,
                   'rdf': sbol3.RDF_XML}
        if ext in ext_map:
            return ext_map[ext]
        else:
            return None

    def test_read_all(self):
        # In lieu of round tripping the files, just make sure we can
        # read them all.
        # This is intended as a temporary test until the library is
        # more complete.
        skip_files = [
            # See https://github.com/SynBioDex/SBOLTestSuite/issues/22
            'annotation',
        ]
        for f in self.find_all_files(SBOL3_LOCATION):
            basename = os.path.basename(f)
            if os.path.splitext(basename)[0] in skip_files:
                # print(f'Skipping {f}')
                continue
            rdf_type = self.rdf_type(f)
            if rdf_type is None:
                # Skip file types we don't know
                # print(f'Skipping {f} of type {rdf_type}')
                continue
            # print(f'Reading {f}')
            doc = sbol3.Document()
            doc.read(f, rdf_type)

    def run_round_trip_file(self, test_path, file_format):
        """Runs a round trip test on the file at the given path.
        Path can be relative or absolute.
        """
        filename = os.path.basename(test_path)
        test2_path = os.path.join(self.temp_out_dir, filename)
        # Read the document, then write it back to disk
        doc = sbol3.Document()
        doc.read(test_path, file_format)
        doc.write(test2_path, file_format)

        # Read the newly written document and compare results
        doc2 = sbol3.Document()
        doc2.read(test2_path, file_format)
        # TODO: what about Document.compare()?
        # self.assertTrue(doc.compare(doc2))

        # Now compare the graphs in RDF
        g1 = rdflib.Graph()
        g1.load(test_path, format=file_format)
        iso1 = rdflib.compare.to_isomorphic(g1)
        g2 = rdflib.Graph()
        g2.load(test2_path, format=file_format)
        iso2 = rdflib.compare.to_isomorphic(g2)
        rdf_diff = rdflib.compare.graph_diff(iso1, iso2)
        if rdf_diff[1] or rdf_diff[2]:
            self.logger.warning('Detected %d different RDF triples in %s' %
                                (len(rdf_diff[1]) + len(rdf_diff[2]), test_path))
            if not self.logger.isEnabledFor(logging.DEBUG):
                self.logger.warning('Set environment variable %s to see details',
                                    DEBUG_ENV_VAR)
            for stmt in rdf_diff[1]:
                self.logger.debug('Only in original: %r', stmt)
            for stmt in rdf_diff[2]:
                self.logger.debug('Only in loaded: %r', stmt)
            self.fail('Differences in RDF detected')

    def test_sbol3_files(self):
        test_dir = SBOL3_LOCATION
        skip_list = [
            # See https://github.com/SynBioDex/SBOLTestSuite/issues/22
            'annotation',
        ]
        for test_file in self.find_all_files(test_dir):
            basename = os.path.basename(test_file)
            if os.path.splitext(basename)[0] in skip_list:
                self.logger.debug(f'Skipping {test_file}')
                continue
            file_format = self.rdf_type(test_file)
            if not file_format:
                continue
            with self.subTest(filename=test_file):
                self.setUp()
                self.run_round_trip_file(test_file, file_format)
                self.tearDown()


if __name__ == '__main__':
    unittest.main()



import os
import unittest

import sbol3

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
SBOL3_LOCATION = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL3')


class TestAttachment(unittest.TestCase):

    def test_create(self):
        base_url = 'https://github.com/SynBioDex/pySBOL3/'
        source_uri = base_url + 'blob/master/sbol3/attachment.py'
        att = sbol3.Attachment('attachment1', source_uri)
        self.assertEqual(source_uri, att.source)

    def test_read_from_file(self):
        test_file = os.path.join(SBOL3_LOCATION, 'entity', 'attachment',
                                 'attachment.nt')
        doc = sbol3.Document()
        doc.read(test_file, sbol3.NTRIPLES)
        search_uri = 'http://parts.igem.org/Part:/attachment1'
        attachment = doc.find(search_uri)
        self.assertIsNotNone(attachment)
        self.assertIsInstance(attachment, sbol3.Attachment)
        self.assertEqual('https://sbolstandard.org/attachment1', attachment.source)
        self.assertEqual('https://identifiers.org/edam:format_2585', attachment.format)
        self.assertEqual(1000, attachment.size)
        self.assertEqual('aaa', attachment.hash)
        self.assertEqual('Alg1', attachment.hash_algorithm)


if __name__ == '__main__':
    unittest.main()



import unittest

import sbol3


class TestVariableComponent(unittest.TestCase):

    def test_create(self):
        vc = sbol3.VariableFeature()
        self.assertIsNotNone(vc)
        # Verify the correct default values
        self.assertEqual(sbol3.SBOL_ZERO_OR_MORE, vc.cardinality)
        self.assertEqual(sbol3.PYSBOL3_MISSING, vc.variable)

    def test_invalid_create(self):
        my_cardinality = 'https://github.com/synbiodex/pysbol3#someNumber'
        with self.assertRaises(sbol3.ValidationError):
            vc = sbol3.VariableFeature(cardinality=my_cardinality)


if __name__ == '__main__':
    unittest.main()



import logging
import os
import subprocess
import sys
import unittest
import tempfile
import shutil
import rdflib
import rdflib.compare

import sbol3

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
EXAMPLE_DIR = os.path.join(os.path.dirname(TEST_DIR), 'examples')
CIRCUIT_EXAMPLE = os.path.join(EXAMPLE_DIR, 'circuit.py')
EXPECTED_CIRCUIT = os.path.join(TEST_DIR, 'resources', 'circuit.nt')


class TestExamples(unittest.TestCase):

    def setUp(self):
        # Create temp directory
        self.temp_out_dir = tempfile.mkdtemp()
        self.logger = logging.getLogger('sbol3.test')
        if not self.logger.hasHandlers():
            logging.basicConfig()

    def tearDown(self):
        # Remove directory after the test
        if self.temp_out_dir:
            shutil.rmtree(self.temp_out_dir)
            self.temp_out_dir = None

    def test_circuit_example(self):
        cmd = [sys.executable, CIRCUIT_EXAMPLE]
        subprocess.check_call(cmd, cwd=self.temp_out_dir)
        out_path = os.path.join(self.temp_out_dir, 'circuit.nt')
        self.assertTrue(os.path.exists(out_path))
        # Load the output
        actual_graph = rdflib.Graph()
        actual_graph.load(out_path, format=sbol3.NTRIPLES)
        actual_iso = rdflib.compare.to_isomorphic(actual_graph)
        # Load the expected output
        expected_graph = rdflib.Graph()
        expected_graph.load(EXPECTED_CIRCUIT, format=sbol3.NTRIPLES)
        expected_iso = rdflib.compare.to_isomorphic(expected_graph)
        rdf_diff = rdflib.compare.graph_diff(expected_iso, actual_iso)
        if rdf_diff[1] or rdf_diff[2]:
            self.logger.warning('Detected %d different RDF triples in %s' %
                                (len(rdf_diff[1]) + len(rdf_diff[2]), out_path))
            for stmt in rdf_diff[1]:
                self.logger.warning('Only in expected: %r', stmt)
            for stmt in rdf_diff[2]:
                self.logger.warning('Only in actual: %r', stmt)
            self.fail('Differences in RDF detected')


if __name__ == '__main__':
    unittest.main()



import math
from typing import List, Union

from . import *


class Component(TopLevel):

    def __init__(self, identity: str, component_type: Union[List[str], str],
                 *, type_uri: str = SBOL_COMPONENT):
        super().__init__(identity, type_uri)
        if isinstance(component_type, str):
            component_type = [component_type]
        self.types: Union[List, Property] = URIProperty(self, SBOL_TYPE, 1, math.inf,
                                                        initial_value=component_type)
        self.roles = URIProperty(self, SBOL_ROLE, 0, math.inf)
        self.sequences = ReferencedObject(self, SBOL_SEQUENCES, 0, math.inf)
        self.features = OwnedObject(self, SBOL_FEATURES, 0, math.inf,
                                    type_constraint=Feature)
        self.interactions = OwnedObject(self, SBOL_INTERACTIONS, 0, math.inf,
                                        type_constraint=Interaction)
        self.constraints = OwnedObject(self, SBOL_CONSTRAINTS, 0, math.inf,
                                       type_constraint=Constraint)
        self.interfaces = OwnedObject(self, SBOL_INTERFACES, 0, 1,
                                      type_constraint=Interface)
        self.models = ReferencedObject(self, SBOL_MODELS, 0, math.inf)
        self.validate()

    def _validate_types(self) -> None:
        # A Component is REQUIRED to have one or more type properties (Section 6.4)
        if len(self.types) < 1:
            message = f'Component {self.identity} has no types'
            raise ValidationError(message)

    def validate(self) -> None:
        super().validate()
        self._validate_types()


def build_component(identity: str, *, type_uri: str = SBOL_COMPONENT) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = Component(identity, [missing], type_uri=type_uri)
    # Remove the dummy values
    obj._properties[SBOL_TYPE] = []
    return obj


Document.register_builder(SBOL_COMPONENT, build_component)



import posixpath
from typing import Union, Any, List, Optional, Dict

import rdflib

from . import *
from .utils import parse_class_name


class OwnedObjectPropertyMixin:

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.type_constraint = None

    def from_user(self, value: Any) -> rdflib.Literal:
        desired_type = SBOLObject
        if self.type_constraint:
            desired_type = self.type_constraint
        if not isinstance(value, desired_type):
            raise TypeError(f'Expecting SBOLObject, got {type(value)}')
        return value

    def to_user(self, value: Any) -> str:
        return value

    def _storage(self) -> Dict[str, list]:
        return self.property_owner._owned_objects

    def item_added(self, item: Any) -> None:
        if not self.property_owner.identity_is_url():
            return
        # if not item.display_id:
        #     raise ValueError(f'Item {item} does not have a display_id')
        type_name = parse_class_name(item.type_uri)
        counter_value = self.property_owner.counter_value(type_name)
        new_display_id = f'{type_name}{counter_value}'
        new_url = posixpath.join(self.property_owner.identity, new_display_id)
        for sibling in self._storage()[self.property_uri]:
            if sibling == item:
                continue
            if sibling.identity == new_url:
                raise ValidationError(f'Duplicate URI: {new_url}')
        item._update_identity(new_url, new_display_id)


class OwnedObjectSingletonProperty(OwnedObjectPropertyMixin, SingletonProperty):

    def __init__(self, property_owner: Any, property_uri: str,
                 lower_bound: int, upper_bound: int,
                 validation_rules: Optional[List] = None,
                 initial_value: Optional[str] = None,
                 type_constraint: Optional[Any] = None):
        super().__init__(property_owner=property_owner,
                         property_uri=property_uri,
                         lower_bound=lower_bound,
                         upper_bound=upper_bound,
                         validation_rules=validation_rules)
        # Better to pass this up to super, but the super().__init__
        # methods are not set up properly for that. They would need
        # to accept unknown keyword arguments.
        self.type_constraint = type_constraint
        if initial_value:
            self.set(initial_value)


class OwnedObjectListProperty(OwnedObjectPropertyMixin, ListProperty):

    def __init__(self, property_owner: Any, property_uri: str,
                 lower_bound: int, upper_bound: int,
                 validation_rules: Optional[List] = None,
                 initial_value: Optional[str] = None,
                 type_constraint: Optional[Any] = None):
        super().__init__(property_owner=property_owner,
                         property_uri=property_uri,
                         lower_bound=lower_bound,
                         upper_bound=upper_bound,
                         validation_rules=validation_rules)
        # Better to pass this up to super, but the super().__init__
        # methods are not set up properly for that. They would need
        # to accept unknown keyword arguments.
        self.type_constraint = type_constraint
        if initial_value:
            self.set(initial_value)


def OwnedObject(property_owner: Any, property_uri: str,
                lower_bound: int, upper_bound: Union[int, float],
                validation_rules: Optional[List] = None,
                initial_value: Optional[Union['Identified', List['Identified']]] = None,
                type_constraint: Optional[Any] = None
                ) -> Property:
    if upper_bound == 1:
        return OwnedObjectSingletonProperty(property_owner, property_uri,
                                            lower_bound, upper_bound,
                                            validation_rules, initial_value,
                                            type_constraint=type_constraint)
    return OwnedObjectListProperty(property_owner, property_uri,
                                   lower_bound, upper_bound,
                                   validation_rules, initial_value,
                                   type_constraint=type_constraint)



import math

from . import *
# Feature is not exported
from .feature import Feature
from .typing import *


class LocalSubComponent(Feature):

    def __init__(self, types: List[str],
                 *, identity: str = None,
                 type_uri: str = SBOL_LOCAL_SUBCOMPONENT) -> None:
        super().__init__(identity, type_uri)
        self.types: uri_list = URIProperty(self, SBOL_TYPE, 1, math.inf,
                                           initial_value=types)
        self.locations = OwnedObject(self, SBOL_LOCATION, 0, math.inf,
                                     type_constraint=Location)
        self.validate()

    def validate(self):
        super().validate()
        if len(self.types) < 1:
            raise ValidationError('LocalSubComponent must have at least 1 type')


def build_local_subcomponent(identity: str,
                             *, type_uri: str = SBOL_LOCAL_SUBCOMPONENT) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = LocalSubComponent([missing], identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[SBOL_TYPE] = []
    return obj


Document.register_builder(SBOL_LOCAL_SUBCOMPONENT, build_local_subcomponent)



import os
import pathlib
import sys
import unittest

try:
    import pycodestyle
except ModuleNotFoundError:
    # The pycodestyle test will be skipped if the module is not available
    pass


MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
# The sbol2 directory is a sibling to the test directory where this file
# lives.
SBOL_PATH = os.path.join(os.path.dirname(MODULE_LOCATION), 'sbol3')
TEST_PATH = MODULE_LOCATION
EXAMPLES_PATH = os.path.join(os.path.dirname(MODULE_LOCATION), 'examples')
SETUP_PATH = os.path.join(os.path.dirname(MODULE_LOCATION), 'setup.py')
STYLE_CONFIG = os.path.join(os.path.dirname(MODULE_LOCATION), 'setup.cfg')

# Please don't increase this number!
MAX_WILDCARD_IMPORTS = 0

# -----------------------------------------------------------------
# Locale Fix
#
# If you get an ascii conversion error, you probably do not have a
# local set.
#
# Follow these steps, substituting for "en_US.utf8" as appropriate:
#
#     apt update
#     apt install locales
#     locale-gen en_US.utf8
#     export LANG=en_US.utf8
#
# -----------------------------------------------------------------


class TestStyle(unittest.TestCase):

    @unittest.skipUnless('pycodestyle' in sys.modules, "pycodestyle not available")
    def test_pep8(self):
        # Test that we conform to PEP-8 via `pycodestyle`
        # Set quiet to `False` to see the style issues
        quiet = True
        style = pycodestyle.StyleGuide(quiet=quiet,
                                       config_file=STYLE_CONFIG)
        style.options.report.start()
        style.input_dir(SBOL_PATH)
        style.input_dir(TEST_PATH)
        style.input_dir(EXAMPLES_PATH)
        style.input_file(SETUP_PATH)
        style.options.report.stop()
        result = style.options.report
        # Please try not to increase the expected number of errors. Please.
        expected_errors = 0
        msg = f'Found {result.total_errors} code style errors (and warnings).'
        self.assertLessEqual(result.total_errors, expected_errors, msg)


if __name__ == '__main__':
    unittest.main()



import math

from . import *
from .typing import *


class Participation(Identified):

    def __init__(self, roles: List[str],
                 participant: Union[SBOLObject, str],
                 *, identity: str = None,
                 type_uri: str = SBOL_PARTCIPATION) -> None:
        super().__init__(identity, type_uri)
        self.roles: uri_list = URIProperty(self, SBOL_ROLE, 1, math.inf,
                                           initial_value=roles)
        self.participant = ReferencedObject(self, SBOL_PARTICIPANT, 1, 1,
                                            initial_value=participant)
        self.validate()

    def validate(self) -> None:
        super().validate()
        if len(self.roles) < 1:
            raise ValidationError('Participation must have at least one role')
        if self.participant is None:
            raise ValidationError('Participation must have a participant')


def build_participation(identity: str,
                        *, type_uri: str = SBOL_PARTCIPATION) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = Participation([missing], missing, identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[SBOL_ROLE] = []
    obj._properties[SBOL_PARTICIPANT] = []
    return obj


Document.register_builder(SBOL_PARTCIPATION, build_participation)



import datetime
from typing import Union, Any, List, Optional

import dateutil.parser
import rdflib

from . import *


class DateTimePropertyMixin:

    def from_user(self, value):
        # None is ok iff upper bound is 1 and lower bound is 0.
        # If upper bound > 1, attribute is a list and None is not a valid list
        # If lower bound > 0, attribute must have a value, so None is unacceptable
        if value is None:
            return None
        if isinstance(value, str):
            value = dateutil.parser.parse(value)
        if not isinstance(value, datetime.datetime):
            raise TypeError(f'Expecting datetime, got {type(value)}')
        return rdflib.Literal(value)

    def to_user(self, value):
        return dateutil.parser.parse(value)


class DateTimeSingletonProperty(DateTimePropertyMixin, SingletonProperty):

    def __init__(self, property_owner: Any, property_uri: str,
                 lower_bound: int, upper_bound: int,
                 validation_rules: Optional[List] = None,
                 initial_value: Optional[str] = None):
        super().__init__(property_owner, property_uri,
                         lower_bound, upper_bound, validation_rules)
        if initial_value is not None:
            self.set(initial_value)


def DateTimeProperty(property_owner: Any, property_uri: str,
                     lower_bound: int, upper_bound: Union[int, float],
                     validation_rules: Optional[List] = None,
                     initial_value: Optional[Union[int, List[int]]] = None) -> Property:
    if upper_bound == 1:
        return DateTimeSingletonProperty(property_owner, property_uri,
                                         lower_bound, upper_bound,
                                         validation_rules, initial_value)
    raise ValueError('Upper bound > 1 not handled yet')



import math

from . import *


class TopLevel(Identified):

    def __init__(self, identity: str, type_uri: str) -> None:
        # Sanity check identity, which is required for a TopLevel
        # More checking on identity happens in Identified, but Identified
        # does not require an identity, only TopLevel does.
        if not identity or not isinstance(identity, str):
            raise ValueError('Identity must be a non-empty string')
        super().__init__(identity, type_uri)
        self.attachments = ReferencedObject(self, SBOL_HAS_ATTACHMENT, 0, math.inf)

    def validate_identity(self) -> None:
        # TODO: See section 5.1 for rules about identity for TopLevel
        super().validate_identity()

    def validate(self) -> None:
        super().validate()



import math

from . import *

# Feature is not exported
from .feature import Feature
from .typing import *


class ExternallyDefined(Feature):
    """The ExternallyDefined class has been introduced so that
    external definitions in databases like ChEBI or UniProt
    can be referenced.
    """

    def __init__(self, types: List[str], definition: str,
                 *, identity: str = None,
                 type_uri: str = SBOL_EXTERNALLY_DEFINED):
        super().__init__(identity, type_uri)
        self.types: uri_list = URIProperty(self, SBOL_TYPE, 1, math.inf,
                                           initial_value=types)
        self.definition: uri_singleton = URIProperty(self, SBOL_DEFINITION, 1, 1,
                                                     initial_value=definition)
        self.validate()

    def validate(self):
        super().validate()
        if len(self.types) < 1:
            raise ValidationError('ExternallyDefined must contain at least 1 type')
        if self.definition is None:
            raise ValidationError('ExternallyDefined must have a definition')


def build_externally_defined(identity: str,
                             *, type_uri: str = SBOL_EXTERNALLY_DEFINED) -> SBOLObject:
    missing = PYSBOL3_MISSING
    obj = ExternallyDefined([missing], missing, identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[SBOL_TYPE] = []
    obj._properties[SBOL_DEFINITION] = []
    return obj


Document.register_builder(SBOL_EXTERNALLY_DEFINED, build_externally_defined)



import os
import unittest

import sbol3

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
SBOL3_LOCATION = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL3')


class TestInterface(unittest.TestCase):

    def test_create(self):
        interface = sbol3.Interface()
        self.assertIsNotNone(interface)

    def test_read_from_file(self):
        test_file = os.path.join(SBOL3_LOCATION, 'entity', 'interface',
                                 'interface.nt')
        doc = sbol3.Document()
        doc.read(test_file, sbol3.NTRIPLES)
        search_uri = 'https://sbolstandard.org/examples/LacI_producer/interface'
        interface = doc.find(search_uri)
        self.assertIsNotNone(interface)
        self.assertIsInstance(interface, sbol3.Interface)
        tetr_uri = 'https://sbolstandard.org/examples/LacI_producer/TetR_protein'
        laci_uri = 'https://sbolstandard.org/examples/LacI_producer/LacI_protein'
        atc_uri = 'https://sbolstandard.org/examples/LacI_producer/aTC'
        iface_input = [tetr_uri, laci_uri]
        self.assertCountEqual(iface_input, interface.input)
        output = [laci_uri]
        self.assertEqual(output, interface.output)
        non_directional = [atc_uri]
        self.assertEqual(non_directional, interface.non_directional)


if __name__ == '__main__':
    unittest.main()



import unittest

import sbol3


class TestSubComponent(unittest.TestCase):

    def test_create(self):
        instance_of = sbol3.Component('comp1', sbol3.SBO_DNA)
        sc1 = sbol3.SubComponent(instance_of)
        self.assertIsNotNone(sc1)
        self.assertEqual(instance_of.identity, sc1.instance_of)
        sc2 = sbol3.SubComponent(instance_of.identity)
        self.assertEqual(instance_of.identity, sc2.instance_of)

    def test_invalid_create(self):
        # SubComponent requires an `instance_of` argument
        with self.assertRaises(sbol3.ValidationError):
            sbol3.SubComponent('')

    def test_external_instance_of(self):
        # See https://github.com/SynBioDex/pySBOL3/issues/136
        c_uri = 'https://synbiohub.example.org/component1'
        m9_media = 'https://synbiohub.example.org/m9-media'
        e_coli = 'https://synbiohub.example.org/e-coli-DH5-alpha'
        c = sbol3.Component(c_uri, sbol3.SBO_DNA)
        sc1 = sbol3.SubComponent(m9_media)
        sc2 = sbol3.SubComponent(e_coli)
        c.features = [sc1, sc2]
        self.assertEqual(len(c.features), 2)
        doc1 = sbol3.Document()
        doc1.add(c)
        doc2 = sbol3.Document()
        doc2.read_string(doc1.write_string(sbol3.NTRIPLES),
                         sbol3.NTRIPLES)
        c2 = doc2.find(c_uri)
        self.assertIsNotNone(c2)
        self.assertEqual(2, len(c2.features))
        self.assertCountEqual([m9_media, e_coli],
                              [sc.instance_of for sc in c2.features])


if __name__ == '__main__':
    unittest.main()



import unittest

import sbol3


class TestCombinatorialDerivation(unittest.TestCase):

    def test_create(self):
        template = 'https://github.com/synbiodex/pysbol3/component'
        cd1 = sbol3.CombinatorialDerivation('cd1', template)
        self.assertEqual(template, cd1.template)
        comp1 = sbol3.Component('comp1', sbol3.SBO_DNA)
        cd2 = sbol3.CombinatorialDerivation('cd2', comp1)
        self.assertEqual(comp1.identity, cd2.template)

    def test_invalid_strategy(self):
        # Test with invalid strategy
        comp1 = sbol3.Component('comp1', sbol3.SBO_DNA)
        cd1 = sbol3.CombinatorialDerivation('cd1', comp1)
        cd1.strategy = sbol3.SBOL_INLINE
        with self.assertRaises(sbol3.ValidationError):
            cd1.validate()


if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------
# Constants to support SBOL 3
#
# See https://sbolstandard.org/data-model-specification/ for the latest version.
# --------------------------------------------------

SBOL_LOGGER_NAME = 'sbol3'

# ----------
# Namespaces
# ----------

SBOL3_NS = 'http://sbols.org/v3#'
SBOL2_NS = 'http://sbols.org/v2#'
SBOL1_NS = 'http://sbols.org/v1#'

# Provenance
PROV_NS = 'http://www.w3.org/ns/prov#'

# Namespace for Chemical Entities of Biological Interest (ChEBI) terms
CHEBI_NS = 'https://identifiers.org/CHEBI:'

# Namespace for Sequence Ontology (SO) terms
SO_NS = "https://identifiers.org/SO:"

# Namespace for Systems Biology Ontology (SBO) terms
SBO_NS = 'https://identifiers.org/SBO:'

# ----------
# SBOL 3 terms
# ----------
SBOL_ATTACHMENT = SBOL3_NS + 'Attachment'
SBOL_BUILT = SBOL3_NS + 'built'
SBOL_CARDINALITY = SBOL3_NS + 'cardinality'
SBOL_COLLECTION = SBOL3_NS + 'Collection'
SBOL_COMBINATORIAL_DERIVATION = SBOL3_NS + 'CombinatorialDerivation'
SBOL_COMPONENT = SBOL3_NS + 'Component'
SBOL_COMPONENT_REFERENCE = SBOL3_NS + 'ComponentReference'
SBOL_CONSTRAINT = SBOL3_NS + 'Constraint'
SBOL_CONSTRAINTS = SBOL3_NS + 'hasConstraint'
SBOL_CUT = SBOL3_NS + 'Cut'
SBOL_DEFINITION = SBOL3_NS + 'definition'
SBOL_DESCRIPTION = SBOL3_NS + 'description'
SBOL_DISPLAY_ID = SBOL3_NS + 'displayId'
SBOL_ELEMENTS = SBOL3_NS + 'elements'
SBOL_ENCODING = SBOL3_NS + 'encoding'
SBOL_END = SBOL3_NS + 'end'
SBOL_ENTIRE_SEQUENCE = SBOL3_NS + 'EntireSequence'
SBOL_EXPERIMENT = SBOL3_NS + 'Experiment'
SBOL_EXPERIMENTAL_DATA = SBOL3_NS + 'ExperimentalData'
SBOL_EXTERNALLY_DEFINED = SBOL3_NS + 'ExternallyDefined'
SBOL_FEATURES = SBOL3_NS + 'hasFeature'
SBOL_FORMAT = SBOL3_NS + 'format'
SBOL_FRAMEWORK = SBOL3_NS + 'framework'
SBOL_HAS_ATTACHMENT = SBOL3_NS + 'hasAttachment'
SBOL_HAS_MEASURE = SBOL3_NS + 'hasMeasure'
SBOL_HASH = SBOL3_NS + 'hash'
SBOL_HASH_ALGORITHM = SBOL3_NS + 'hashAlgorithm'
SBOL_IDENTIFIED = SBOL3_NS + 'Identified'
SBOL_IMPLEMENTATION = SBOL3_NS + 'Implementation'
SBOL_IN_CHILD_OF = SBOL3_NS + 'inChildOf'
SBOL_INPUT = SBOL3_NS + 'input'
SBOL_INSTANCE_OF = SBOL3_NS + 'instanceOf'
SBOL_INTERACTIONS = SBOL3_NS + 'hasInteraction'
SBOL_INTERACTION = SBOL3_NS + 'Interaction'
SBOL_INTERFACE = SBOL3_NS + 'Interface'
SBOL_INTERFACES = SBOL3_NS + 'hasInterface'
SBOL_LANGUAGE = SBOL3_NS + 'language'
SBOL_LOCAL_SUBCOMPONENT = SBOL3_NS + 'LocalSubComponent'
SBOL_LOCATION = SBOL3_NS + 'hasLocation'
SBOL_MEMBER = SBOL3_NS + 'member'
SBOL_MODEL = SBOL3_NS + 'Model'
SBOL_MODELS = SBOL3_NS + 'hasModel'
SBOL_NAME = SBOL3_NS + 'name'
SBOL_NAMESPACE = SBOL3_NS + 'Namespace'
SBOL_NON_DIRECTIONAL = SBOL3_NS + 'nondirectional'
SBOL_OBJECT = SBOL3_NS + 'object'
SBOL_ORDER = SBOL3_NS + 'order'
SBOL_ORIENTATION = SBOL3_NS + 'orientation'
SBOL_OUTPUT = SBOL3_NS + 'output'
SBOL_PARTICIPANT = SBOL3_NS + 'participant'
SBOL_PARTCIPATION = SBOL3_NS + 'Participation'
SBOL_PARTICIPATIONS = SBOL3_NS + 'hasParticipation'
SBOL_RANGE = SBOL3_NS + 'Range'
SBOL_RESTRICTION = SBOL3_NS + 'restriction'
SBOL_ROLE = SBOL3_NS + 'role'
SBOL_SEQUENCE = SBOL3_NS + 'Sequence'
SBOL_SEQUENCE_FEATURE = SBOL3_NS + 'SequenceFeature'
SBOL_SEQUENCES = SBOL3_NS + 'hasSequence'
SBOL_SIZE = SBOL3_NS + 'size'
SBOL_SOURCE = SBOL3_NS + 'source'
SBOL_SOURCE_LOCATION = SBOL3_NS + 'sourceLocation'
SBOL_START = SBOL3_NS + 'start'
SBOL_STRATEGY = SBOL3_NS + 'strategy'
SBOL_SUBCOMPONENT = SBOL3_NS + 'SubComponent'
SBOL_SUBJECT = SBOL3_NS + 'subject'
SBOL_TEMPLATE = SBOL3_NS + 'template'
SBOL_TOP_LEVEL = SBOL3_NS + 'TopLevel'
SBOL_TYPE = SBOL3_NS + 'type'
SBOL_VARIABLE = SBOL3_NS + 'variable'
SBOL_VARIABLE_FEATURE = SBOL3_NS + 'VariableFeature'
SBOL_VARIABLE_FEATURES = SBOL3_NS + 'hasVariableFeature'
SBOL_VARIANT = SBOL3_NS + 'variant'
SBOL_VARIANT_COLLECTION = SBOL3_NS + 'variantCollection'
SBOL_VARIANT_DERIVATION = SBOL3_NS + 'variantDerivation'
SBOL_VARIANT_MEASURE = SBOL3_NS + 'variantMeasure'

# Recommended values for Sequence encoding
SBOL_IUPAC_DNA = SBOL3_NS + 'iupacNucleicAcid'
SBOL_IUPAC_RNA = SBOL_IUPAC_DNA
SBOL_IUPAC_PROTEIN = SBOL3_NS + 'iupacAminoAcid'
SMILES_ENCODING = 'http://www.opensmiles.org/opensmiles.html'

# Valid values for Feature orientation
# See SBOL3 Section 6.4.1 Table 5
SBOL_INLINE = SBOL3_NS + 'inline'
SBOL_REVERSE_COMPLEMENT = SBOL3_NS + 'reverseComplement'

# Valid values for Combinatorial Derivation strategy
# See SBOL3 Section 6.5 Table 12
SBOL_ENUMERATE = SBOL3_NS + 'enumerate'
SBOL_SAMPLE = SBOL3_NS + 'sample'

# Valid values for Variable Feature cardinality
# See SBOL3 Section 6.5 Table 13
SBOL_ONE = SBOL3_NS + 'one'
SBOL_ONE_OR_MORE = SBOL3_NS + 'oneOrMore'
SBOL_ZERO_OR_MORE = SBOL3_NS + 'zeroOrMore'
SBOL_ZERO_OR_ONE = SBOL3_NS + 'zeroOrOne'

# A namespace for internal URIs and for testing
PYSBOL3_NS = 'https://github.com/synbiodex/pysbol3#'
PYSBOL3_MISSING = PYSBOL3_NS + 'missing'


# ----------
# Provenance terms
# ----------
PROV_ACTIVITY = PROV_NS + 'Activity'
PROV_AGENT = PROV_NS + 'Agent'
PROV_AGENTS = PROV_NS + 'agent'
PROV_ASSOCIATION = PROV_NS + 'Association'
PROV_DERIVED_FROM = PROV_NS + 'wasDerivedFrom'
PROV_ENTITY = PROV_NS + 'entity'
PROV_GENERATED_BY = PROV_NS + 'wasGeneratedBy'
PROV_PLAN = PROV_NS + 'Plan'
PROV_PLANS = PROV_NS + 'hadPlan'
PROV_QUALIFIED_ASSOCIATION = PROV_NS + 'qualifiedAssociation'
PROV_QUALIFIED_USAGE = PROV_NS + 'qualifiedUsage'
PROV_ROLES = PROV_NS + 'hadRole'
PROV_STARTED_AT_TIME = PROV_NS + 'startedAtTime'
PROV_ENDED_AT_TIME = PROV_NS + 'endedAtTime'
PROV_USAGE = PROV_NS + 'Usage'


# ----------
# Component/Feature roles
#
# * These are common, others can be used as well.
# * See the SBOL 3 spec, Section 6.4, Table 4
# ----------
SO_PROMOTER = SO_NS + "0000167"
SO_RBS = SO_NS + "0000139"
SO_CDS = SO_NS + "0000316"
SO_TERMINATOR = SO_NS + "0000141"
SO_GENE = SO_NS + "0000704"
SO_OPERATOR = SO_NS + "0000057"
SO_ENGINEERED_GENE = SO_NS + "0000280"
SO_ENGINEERED_REGION = SO_NS + '0000804'
SO_MRNA = SO_NS + "0000234"
CHEBI_EFFECTOR = CHEBI_NS + '35224'
SO_TRANSCRIPTION_FACTOR = SO_NS + "0003700"

# Component types
# See the SBOL 3 spec, Section 6.4, Table 2
SBO_DNA = SBO_NS + '0000251'
SBO_RNA = SBO_NS + '0000250'
SBO_PROTEIN = SBO_NS + '0000252'
SBO_SIMPLE_CHEMICAL = SBO_NS + '0000247'
SBO_NON_COVALENT_COMPLEX = SBO_NS + '0000253'
SBO_FUNCTIONAL_ENTITY = SBO_NS + '0000241'

# Component types
# See the SBOL 3 spec, Section 6.4, Table 3
SO_LINEAR = SO_NS + "0000987"
SO_CIRCULAR = SO_NS + "0000988"
SO_SINGLE_STRANDED = SO_NS + "0000984"
SO_DOUBLE_STRANDED = SO_NS + "0000985"

# RECOMMENDED URIs for expressing identity and orientation with
# the restriction property.
# See the SBOL 3 spec, Section 6.4, Table 7
SBOL_VERIFY_IDENTICAL = SBOL3_NS + 'verifyIdentical'
SBOL_DIFFERENT_FROM = SBOL3_NS + 'differentFrom'
SBOL_REPLACES = SBOL3_NS + 'replaces'
SBOL_SAME_ORIENTATION_AS = SBOL3_NS + 'sameOrientationAs'
SBOL_OPPOSITE_ORIENTATION_AS = SBOL3_NS + 'oppositeOrientationAs'

# RECOMMENDED URIs for expressing topological relations with
# the restriction property.
# See the SBOL 3 spec, Section 6.4, Table 8
SBOL_IS_DISJOINT_FROM = SBOL3_NS + 'isDisjointFrom'
SBOL_STRICTLY_CONTAINS = SBOL3_NS + 'strictlyContains'
SBOL_CONTAINS = SBOL3_NS + 'contains'
SBOL_EQUALS = SBOL3_NS + 'equals'
SBOL_MEETS = SBOL3_NS + 'meets'
SBOL_COVERS = SBOL3_NS + 'covers'
SBOL_OVERLAPS = SBOL3_NS + 'overlaps'

# RECOMMENDED URIs for expressing sequential relations with the
# restriction property
# See the SBOL 3 spec, Section 6.4, Table 9
SBOL_PRECEDES = SBOL3_NS + 'precedes'
SBOL_STRICTLY_PRECEDES = SBOL3_NS + 'strictlyPrecedes'
# SBOL_MEETS = SBOL3_NS + 'meets'
# SBOL_OVERLAPS = SBOL3_NS + 'overlaps'
# SBOL_CONTAINS = SBOL3_NS + 'contains'
# SBOL_STRICTLY_CONTAINS = SBOL3_NS + 'strictlyContains'
# SBOL_EQUALS = SBOL3_NS + 'equals'
SBOL_FINISHES = SBOL3_NS + 'finishes'
SBOL_STARTS = SBOL3_NS + 'starts'

# Interaction types
# See the SBOL 3 spec, Section 6.4, Table 10
SBO_INHIBITION = SBO_NS + '0000169'
SBO_STIMULATION = SBO_NS + '0000170'
SBO_BIOCHEMICAL_REACTION = SBO_NS + '0000176'
SBO_NON_COVALENT_BINDING = SBO_NS + '0000177'
SBO_DEGRADATION = SBO_NS + '0000179'
SBO_GENETIC_PRODUCTION = SBO_NS + '0000589'
SBO_CONTROL = SBO_NS + '0000168'

# Participation roles
# See the SBOL 3 spec, Section 6.4, Table 11
SBO_INHIBITOR = SBO_NS + '0000020'
SBO_INHIBITED = SBO_NS + '0000642'
SBO_STIMULATOR = SBO_NS + '0000459'
SBO_STIMULATED = SBO_NS + '0000643'
SBO_REACTANT = SBO_NS + '0000010'
SBO_PRODUCT = SBO_NS + '0000011'
SBO_PROMOTER = SBO_NS + '0000598'
SBO_MODIFIER = SBO_NS + '0000019'
SBO_MODIFIED = SBO_NS + '0000644'
SBO_TEMPLATE = SBO_NS + '0000645'

# RDF File Formats
NTRIPLES = 'nt'
RDF_XML = 'xml'
TURTLE = 'ttl'
JSONLD = 'json-ld'
SORTED_NTRIPLES = 'sorted nt'

# Ontology of Units of Measure
OM_NS = 'http://www.ontology-of-units-of-measure.org/resource/om-2/'

OM_ALTERNATIVE_LABEL = OM_NS + 'alternativeLabel'
OM_ALTERNATIVE_SYMBOL = OM_NS + 'alternativeSymbol'
OM_BINARY_PREFIX = OM_NS + 'BinaryPrefix'
# TODO: this should really be http://www.w3.org/2000/01/rdf-schema#comment
OM_COMMENT = OM_NS + 'comment'
OM_HAS_BASE = OM_NS + 'hasBase'
OM_HAS_DENOMINATOR = OM_NS + 'hasDenominator'
OM_HAS_EXPONENT = OM_NS + 'hasExponent'
OM_HAS_FACTOR = OM_NS + 'hasFactor'
OM_HAS_NUMERATOR = OM_NS + 'hasNumerator'
OM_HAS_NUMERICAL_VALUE = OM_NS + 'hasNumericalValue'
OM_HAS_PREFIX = OM_NS + 'hasPrefix'
OM_HAS_TERM1 = OM_NS + 'hasTerm1'
OM_HAS_TERM2 = OM_NS + 'hasTerm2'
OM_HAS_UNIT = OM_NS + 'hasUnit'
# TODO: this should really be http://www.w3.org/2000/01/rdf-schema#label
OM_LABEL = OM_NS + 'label'
OM_LONG_COMMENT = OM_NS + 'longcomment'
OM_MEASURE = OM_NS + 'Measure'
OM_PREFIXED_UNIT = OM_NS + 'PrefixedUnit'
OM_SINGULAR_UNIT = OM_NS + 'SingularUnit'
OM_SI_PREFIX = OM_NS + 'SIPrefix'
OM_SYMBOL = OM_NS + 'symbol'
OM_UNIT_DIVISION = OM_NS + 'UnitDivision'
OM_UNIT_EXPONENTIATION = OM_NS + 'UnitExponentiation'
OM_UNIT_MULTIPLICATION = OM_NS + 'UnitMultiplication'



import os
import unittest

import sbol3

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
SBOL3_LOCATION = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL3')


class TestComponentReference(unittest.TestCase):

    def test_create(self):
        in_child_of = 'https://github.com/synbiodex/pysbol3/subcomponent'
        feature = 'https://github.com/synbiodex/pysbol3/other_feature'
        comp_ref = sbol3.ComponentReference(in_child_of, feature)
        self.assertIsNotNone(comp_ref)
        self.assertEqual(in_child_of, comp_ref.in_child_of)
        self.assertEqual(feature, comp_ref.feature)

    def test_read_from_file(self):
        test_file = os.path.join(SBOL3_LOCATION, 'toggle_switch',
                                 'toggle_switch.nt')
        doc = sbol3.Document()
        doc.read(test_file, sbol3.NTRIPLES)
        uri = 'https://sbolstandard.org/examples/toggle_switch/componentreference_1'
        comp_ref = doc.find(uri)
        self.assertIsNotNone(comp_ref)
        self.assertIsInstance(comp_ref, sbol3.ComponentReference)
        in_child_of = 'https://sbolstandard.org/examples/toggle_switch/LacI_producer'
        self.assertEqual(in_child_of, comp_ref.in_child_of)
        feature = 'https://sbolstandard.org/examples/LacI_producer/LacI_protein'
        self.assertEqual(feature, comp_ref.feature)


if __name__ == '__main__':
    unittest.main()



import abc
from typing import Union

from . import *

int_property = Union[IntProperty, int]


class Location(Identified, abc.ABC):

    def __init__(self, seq_or_uri: Union[Sequence, str],
                 identity: str, type_uri: str) -> None:
        super().__init__(identity, type_uri)
        self.orientation = URIProperty(self, SBOL_ORIENTATION, 0, 1)
        self.order = IntProperty(self, SBOL_ORDER, 0, 1)
        self.sequence = ReferencedObject(self, SBOL_SEQUENCES, 1, 1,
                                         initial_value=seq_or_uri)

    def validate(self) -> None:
        super().validate()
        if not self.sequence:
            raise ValidationError(f'Location {self.identity} does not have a sequence')


class Range(Location):

    def __init__(self, seq_or_uri: Union[Sequence, str], start: int, end: int,
                 *, identity: str = None, type_uri: str = SBOL_RANGE) -> None:
        super().__init__(seq_or_uri, identity, type_uri)
        self.start: int_property = IntProperty(self, SBOL_START, 1, 1,
                                               initial_value=start)
        self.end: int_property = IntProperty(self, SBOL_END, 1, 1,
                                             initial_value=end)
        self.validate()

    def validate(self) -> None:
        super().validate()
        if self.start < 1:
            raise ValidationError('Start must be greater than 0')
        if self.end < 1:
            raise ValidationError('Start must be greater than 0')
        if self.end < self.start:
            raise ValidationError('End must be >= start')


def build_range(identity: str, type_uri: str = SBOL_RANGE):
    """Used by Document to construct a Range when reading an SBOL file.
    """
    start = 1
    end = 1
    obj = Range(PYSBOL3_MISSING, start, end, identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[SBOL_SEQUENCES] = []
    obj._properties[SBOL_START] = []
    obj._properties[SBOL_END] = []
    return obj


Document.register_builder(SBOL_RANGE, build_range)


class Cut(Location):

    def __init__(self, seq_or_uri: Union[Sequence, str], at: int,
                 *, identity: str = None, type_uri: str = SBOL_CUT) -> None:
        super().__init__(seq_or_uri, identity, type_uri)
        self.at: int_property = IntProperty(self, SBOL_START, 1, 1,
                                            initial_value=at)
        self.validate()

    def validate(self) -> None:
        super().validate()
        if self.at < 0:
            raise ValidationError('At must be >= 0')


def build_cut(identity: str, type_uri: str = SBOL_CUT):
    """Used by Document to construct a Cut when reading an SBOL file.
    """
    at = 0
    obj = Cut(PYSBOL3_MISSING, at, identity=identity, type_uri=type_uri)
    # Remove the dummy values
    obj._properties[SBOL_SEQUENCES] = []
    obj._properties[SBOL_START] = []
    return obj


Document.register_builder(SBOL_CUT, build_cut)


class EntireSequence(Location):

    def __init__(self, seq_or_uri: Union[Sequence, str],
                 *, identity: str = None,
                 type_uri: str = SBOL_ENTIRE_SEQUENCE) -> None:
        super().__init__(seq_or_uri, identity, type_uri)
        self.validate()


def build_entire_sequence(identity: str, type_uri: str = SBOL_ENTIRE_SEQUENCE):
    """Used by Document to construct an EntireSequence when reading an SBOL file.
    """
    obj = EntireSequence(PYSBOL3_MISSING, identity=identity, type_uri=type_uri)
    obj._properties[SBOL_SEQUENCES] = []
    return obj


Document.register_builder(SBOL_ENTIRE_SEQUENCE, EntireSequence)



import os
import tempfile
import unittest

import sbol3


class TestAnnotation(unittest.TestCase):

    def test_annotation(self):
        # Create custom annotation
        annotation_uri = 'http://example.org/boolean_property'
        annotation_value = 'foo'

        c = sbol3.Component('c1', sbol3.SBO_DNA)
        c.annotation = sbol3.TextProperty(c, annotation_uri,
                                          0, 1, [])
        c.annotation = annotation_value
        self.assertEqual(annotation_value, c.annotation)

        doc = sbol3.Document()
        doc.add(c)
        doc2 = sbol3.Document()
        with tempfile.TemporaryDirectory() as tmpdirname:
            test_file = os.path.join(tmpdirname, 'annotation.xml')
            doc.write(test_file, sbol3.RDF_XML)
            # Roundtrip
            doc2.read(test_file, sbol3.RDF_XML)

        # Recover annotation
        c = doc2.find('c1')
        c.annotation = sbol3.TextProperty(c, annotation_uri,
                                          0, 1, [])
        self.assertEqual(annotation_value, c.annotation)


if __name__ == '__main__':
    unittest.main()
